import logging
import datetime
import json
from time import sleep
from nubium_utils.custom_exceptions import NoMessageError, ConsumeMessageError
from nubium_utils.metrics import MetricsManager
from .confluent_runtime_vars import env_vars


LOGGER = logging.getLogger(__name__)


def _wait_until_message_time(msg_timestamp, guid):
    """
    Wait until the message's timestamp + the deployments offset before handling
    """
    wait_minutes = int(env_vars()['NU_CONSUMER_TIMESTAMP_OFFSET_MINUTES'])
    if wait_minutes:
        message_process_time = (msg_timestamp // 1000) + (wait_minutes * 60)

        wait_time = message_process_time - datetime.datetime.timestamp(datetime.datetime.utcnow())

        if wait_time > 0:
            LOGGER.info(f'Waiting {wait_time} seconds before retry message processing continues; GUID {guid}')
            sleep(wait_time)
            return 'i sleep'
    return 'real shit'


def poll_for_message(consumer, timeout=None):
    """
    Polls the broker for a message using the given timeout.
    If there are no messages to consume, either because None is returned
    or the message error is the no messages error,
    raises a NoMessageError.
    """
    message = consumer.poll(timeout if timeout else int(env_vars()['NU_CONSUMER_POLL_TIMEOUT']))
    if message is None:
        raise NoMessageError
    return message


def handle_consumed_message(message, monitor=None):
    """
    Handles a consumed message to check for errors, handle retry waits, and log the consumption as a metric

    If the message is returned with a breaking error,
    raises a ConsumeMessageError.

    If the message is valid, waits until the message's timestamp plus
    the current process's time offset before handling the message.
    This allows retry deployments to wait in a non-
    blocking fashion
    """
    guid = None
    try:
        guid = [item[1] for item in message.headers() if item[0] == 'guid'][0].decode()
        if monitor:
            monitor.set_seconds_behind(round(datetime.datetime.timestamp(datetime.datetime.utcnow())) - message.timestamp()[1]//1000)
        if '__changelog' not in message.topic():
            LOGGER.info(f"Message consumed from topic {message.topic()} partition {message.partition()}, offset {message.offset()}; GUID {guid}")
            LOGGER.debug(f"Consumed message key: {repr(message.key())}")
    except AttributeError:
        if "object has no attribute 'headers'" in str(message.error()):
            raise ConsumeMessageError("Headers were inaccessible on the message. Potentially a corrupt message?")
    except IndexError:
        raise ConsumeMessageError("no GUID was provided in the message headers.")

    # Wait until message time if using a retry process
    _wait_until_message_time(message.timestamp()[1], guid)

    # Increment the metric for consumed messages by one
    if monitor:
        monitor.inc_messages_consumed(1, message.topic())


def consume_message(consumer, monitor: MetricsManager, timeout=None):
    """
    Consumes a message from the broker while handling errors and waiting if necessary

    If the message is valid, then the message is returned
    """
    message = poll_for_message(consumer, timeout)
    handle_consumed_message(message, monitor)
    return message



import logging
from .confluent_runtime_vars import env_vars
from .producer_utils import confirm_produce

LOGGER = logging.getLogger(__name__)


def handle_no_messages(no_msg_exception=None, producers=None):
    """
    Since producer.poll() is typically called within the NU produce_message method, we want to acknowledge
    any outstanding produce attempts while the app has nothing to consume (and thus not actively producing/polling).
    """
    if no_msg_exception:
        LOGGER.debug(no_msg_exception)
    if producers:
        LOGGER.debug('flushing remaining producer queues')
        confirm_produce(producers)


def shutdown_cleanup(producers=None, consumer=None):
    """
    As part of shutdown, make sure all queued up produce messages are flushed, gracefully kill the consumer.
    """
    LOGGER.info('Performing graceful teardown of producer and/or consumer...')
    if consumer:
        LOGGER.debug("Shutting down consumer; no further commits can be queued or finalized.")
        consumer.close()
    if producers:
        LOGGER.debug("Sending/confirming the leftover messages in producer message queue")
        confirm_produce(producers, timeout=30)



import logging
from os import environ
from time import sleep

import pytest
from python_eloqua_wrapper import EloquaSession

LOGGER = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def eloqua_session():
    yield EloquaSession(
        username=environ['ELOQUA_USER'],
        password=environ['ELOQUA_PASSWORD'],
        company=environ['ELOQUA_COMPANY'],
    )


@pytest.fixture()
def eloqua_bulk_wait():
    LOGGER.info("Waiting to ensure Eloqua had time to update the records via BULK")
    sleep(20)


@pytest.fixture()
def create_delete_contacts(map_external_input, eloqua_session):
    LOGGER.info("Creating contacts in Eloqua...")
    contact_ids = []
    for contact in map_external_input:
        response = eloqua_session.post(url='/api/REST/1.0/data/contact', json=contact)
        response.raise_for_status()
        contact.update({'id': response.json().get('id')})
        contact_id = response.json().get('id')
        contact_ids.append(contact_id)
    yield contact_ids
    LOGGER.info("Deleting contacts from Eloqua...")
    for contact_id in contact_ids:
        response = eloqua_session.delete(url=f'/api/REST/1.0/data/contact/{contact_id}')
        response.raise_for_status()


@pytest.fixture()
def contact_field_map(eloqua_session):
    response = eloqua_session.get('/api/REST/1.0/assets/contact/fields?depth=complete', timeout=30)
    response.raise_for_status()
    return response.json()['elements']


@pytest.fixture()
def create_contact_field_map(contact_field_map):
    return {
        contact_field['internalName']: contact_field['id']
        for contact_field in contact_field_map}


@pytest.fixture()
def retrieved_contact_field_map(contact_field_map):
    result = {contact_field['id']: contact_field['internalName'] for contact_field in contact_field_map}
    return result


def map_retrieved_contact_record(record, retrieved_contact_field_map):
    field_map = {retrieved_contact_field_map[field['id']]: field.get('value', "") for field in record['fieldValues']}
    field_map['C_EmailAddress'] = record.get("emailAddress", "")
    field_map['C_FirstName'] = record.get("firstName", "")
    field_map['C_LastName'] = record.get("lastName", "")
    field_map['C_MobilePhone'] = record.get("mobilePhone", "")
    field_map['C_Country'] = record.get("country", "")
    field_map['C_Address1'] = record.get("address1", "")
    field_map['C_Address2'] = record.get("address2", "")
    field_map['C_Address3'] = record.get("address3", "")
    field_map['C_City'] = record.get("city", "")
    field_map['C_State_Prov'] = record.get("province", "")
    field_map['C_Zip_Postal'] = record.get("postalCode", "")
    field_map['C_Company'] = record.get("accountName", "")
    field_map['C_BusPhone'] = record.get("businessPhone", "")
    field_map['C_Title'] = record.get("title", "")
    field_map['isBounced'] = record.get("isBounceback", ""),
    field_map['isSubscribed'] = record.get("isSubscribed", "")
    return field_map


@pytest.fixture()
def retrieve_contacts_for_validation(create_delete_contacts, retrieved_contact_field_map, eloqua_session):
    contacts = []
    for contact_id in create_delete_contacts:
        response = eloqua_session.get(url=f'/api/REST/1.0/data/contact/{contact_id}')
        response.raise_for_status()
        contacts.append(map_retrieved_contact_record(response.json(), retrieved_contact_field_map))
    return contacts



import pytest
from unittest.mock import patch, MagicMock, call
import os
from nubium_utils.confluent_utils.consumer_utils import *
from nubium_utils.confluent_utils.consumer_utils import _wait_until_message_time
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.test_utils import MockMessage
from datetime import datetime, timedelta
import time_machine
from pytz import timezone


def required_env_vars():
    return {
        "NU_LOGLEVEL": "DEBUG",
        "NU_HOSTNAME": "test-app-0",
        "NU_APP_NAME": "test_app",
        "NU_MP_PROJECT": "none",
        "NU_MP_CLUSTER": "none",
        "NU_DO_METRICS_PUSHING": "false",
        "NU_SCHEMA_REGISTRY_URL": "sr_url",
        "NU_SCHEMA_REGISTRY_USERNAME": "sr_un",
        "NU_SCHEMA_REGISTRY_PASSWORD": "sr_pw",
        "NU_TOPIC_CONFIGS_YAML": '{"test_topic_0": {"cluster": "cluster_0"}, "test_topic_1": {"cluster": "cluster_1"}}',
        "NU_KAFKA_CLUSTERS_CONFIGS_YAML": '{"cluster_0": {"url": "www", "username": "un", "password": "pw"}}',
        "NU_CONSUMER_POLL_TIMEOUT": "5",
        "NU_CONSUMER_TIMESTAMP_OFFSET_MINUTES": "1"
    }


def utcnow(seconds=0):
    return datetime(2022, 1, 21, 0, 1, tzinfo=timezone("Etc/UTC")) + timedelta(seconds=seconds)


@pytest.fixture(autouse=True)
def set_env_vars():
    evars = required_env_vars()
    env_patch = patch.dict('os.environ', evars)
    env_patch.start()
    env_vars._reload()


@pytest.fixture
def message(request):
    return MockMessage(**request.param)


@pytest.fixture
def consumer():
    return MagicMock()


@pytest.fixture
def monitor():
    return MagicMock()


class TestConsumerUtils:

    @pytest.mark.skip(reason="issues with time machine unit tests on anything but tim's machine. Otherwise works as expected.")
    @time_machine.travel(utcnow(seconds=-1), tick=False)
    def test__wait_until_message_time__sleep(self):
        assert _wait_until_message_time(1642741200000, '12345') == 'i sleep'

    @pytest.mark.skip(reason="issues with time machine unit tests on anything but tim's machine. Otherwise works as expected.")
    @time_machine.travel(utcnow(seconds=0), tick=False)
    def test__wait_until_message_time__nosleep(self):
        assert _wait_until_message_time(1642741200000, '12345') == 'real shit'

    @pytest.mark.parametrize('message', [{}], indirect=True)
    def test_poll_for_message(self, consumer, message):
        consumer.poll.side_effect = [message, None]
        assert message == poll_for_message(consumer)
        with pytest.raises(NoMessageError):
            poll_for_message(consumer)

    @pytest.mark.skip(reason="issues with time machine unit tests on anything but tim's machine. Otherwise works as expected.")
    @time_machine.travel(utcnow(), tick=False)
    @pytest.mark.parametrize('message', [{"headers": [("test_header_key", b"test_header_value"), ("guid", b"12345")]}], indirect=True)
    def test_handle_consumed_message(self, message, monitor):
        with patch('nubium_utils.confluent_utils.consumer_utils._wait_until_message_time') as wait_patch:
            handle_consumed_message(message)
            monitor.set_seconds_behind.assert_not_called()
            monitor.inc_messages_consumed.assert_not_called()

            handle_consumed_message(message, monitor)
            monitor.set_seconds_behind.assert_called_with(60)
            monitor.inc_messages_consumed.assert_called_with(1, message.topic())

    @pytest.mark.parametrize('message', [{"headers": [], "error": "object has no attribute 'headers'"}], indirect=True)
    def test_handle_consumed_message__missing_headers(self, message):
        with pytest.raises(ConsumeMessageError):
            handle_consumed_message(message)

    @pytest.mark.parametrize('message', [{"headers": [("test_header_key", b"test_header_value")]}], indirect=True)
    def test_handle_consumed_message__no_guid_field(self, message):
        with pytest.raises(ConsumeMessageError):
            handle_consumed_message(message)


if __name__ == '__main__':
    pytest.main([__file__])



import pytest
from unittest.mock import patch, MagicMock, call
import os
from nubium_utils.confluent_utils.gtfo.gtfo_app import *

patch_path = 'nubium_utils.confluent_utils.gtfo.gtfo_app'


@pytest.fixture
def cluster_name():
    return "cluster_0"


@pytest.fixture
def test_topic():
    return "test_topic_0"


@pytest.fixture
def topic_schema_dict(test_topic):
    return {test_topic: {"type": "string"}}


@pytest.fixture
def topic_configs_yaml(cluster_name, test_topic):
    # can be a json as well
    return str({test_topic: {"cluster": cluster_name}, "test_topic_1": {"cluster": cluster_name}})


@pytest.fixture
def cluster_configs_yaml(cluster_name):
    return str({cluster_name: {"url": "www", "username": "un", "password": "pw"}})


@pytest.fixture
def health_filepath(tmp_path_factory):
    return tmp_path_factory.mktemp('health')


@pytest.fixture
def required_env_vars(topic_configs_yaml, cluster_configs_yaml):
    return {
        "NU_LOGLEVEL": "DEBUG",
        "NU_HOSTNAME": "test-app-0",
        "NU_APP_NAME": "test_app",
        "NU_MP_PROJECT": "none",
        "NU_MP_CLUSTER": "none",
        "NU_DO_METRICS_PUSHING": "false",
        "NU_SCHEMA_REGISTRY_URL": "sr_url",
        "NU_SCHEMA_REGISTRY_USERNAME": "sr_un",
        "NU_SCHEMA_REGISTRY_PASSWORD": "sr_pw",
        "NU_TOPIC_CONFIGS_YAML": topic_configs_yaml,
        "NU_KAFKA_CLUSTERS_CONFIGS_YAML": cluster_configs_yaml,
        "NU_CONSUMER_POLL_TIMEOUT": "5",
        "NU_CONSUMER_TIMESTAMP_OFFSET_MINUTES": "1"
    }


@pytest.fixture(autouse=True)
def set_env_vars(required_env_vars):
    evars = required_env_vars
    env_patch = patch.dict('os.environ', evars)
    env_patch.start()
    env_vars._reload()


class MockTransaction(Transaction):
    def __init__(self, **kwargs):
        kwargs_out = dict(producer=MagicMock(), consumer=MagicMock(), message=MagicMock(), metrics_manager=MagicMock())
        kwargs_out.update(**kwargs)
        super().__init__(**kwargs_out)


class MockGtfoApp(GtfoApp):
    def __init__(self, **kwargs):
        kwargs_out = dict(
            app_function=MagicMock(), consume_topics_list=['test_topic_0'], produce_topic_schema_dict={"test_topic_1": {"type": "string"}}, transaction_type=MockTransaction,
            app_function_arglist=['arg_a', 'arg_b'], metrics_manager=MagicMock(), schema_registry=MagicMock(), cluster_name=None, consumer=MagicMock(), producer=MagicMock())
        kwargs_out.update(**kwargs)
        super().__init__(**kwargs_out)


class TestTransaction:
    def test_original_message_functions(self):
        transaction = MockTransaction()
        transaction.key()
        transaction.message.key.assert_called()
        transaction.value()
        transaction.message.value.assert_called()
        transaction.topic()
        transaction.message.topic.assert_called()
        transaction.partition()
        transaction.message.partition.assert_called()
        transaction.offset()
        transaction.message.offset.assert_called()
        
    def test_headers(self):
        transaction = MockTransaction()
        with patch(f'{patch_path}.parse_headers') as patch_head:
            transaction.headers()
            transaction.message.headers.assert_called()
            patch_head.assert_called_with(transaction.message.headers())

    def test_auto_consume(self):
        with patch.object(MockTransaction, 'consume') as patch_consume:
            MockTransaction()
            patch_consume.assert_not_called()

            MockTransaction(auto_consume=False, message=None)
            patch_consume.assert_not_called()

            MockTransaction(message=None)
            patch_consume.assert_called()

    def test_messages(self):
        transaction = MockTransaction()
        assert transaction.messages() == transaction.message

    def test__init_transaction(self):
        transaction = MockTransaction()
        transaction._init_transaction()
        transaction.producer.begin_transaction.assert_called()
        assert transaction._active_transaction

        transaction = MockTransaction()
        transaction._active_transaction = True
        transaction.producer.begin_transaction.assert_not_called()

        transaction = MockTransaction()
        transaction._committed = True
        transaction.producer.begin_transaction.assert_not_called()

    def test_produce(self):
        transaction = MockTransaction()
        with patch.object(MockTransaction, '_init_transaction') as patch_trans:
            with patch(f'{patch_path}.parse_headers') as patch_head:
                patch_head.return_value = {'header': 'value'}
                with patch(f'{patch_path}.produce_message') as patch_produce:
                    transaction.produce({"test_key": "test_value"})
                    patch_produce.assert_called_with(transaction.producer, {"test_key": "test_value"}, transaction.metrics_manager, {'header': 'value'})
                patch_head.assert_called()
            patch_trans.assert_called()
        transaction.producer.poll.assert_has_calls([call(0), call(0)])

    def test_commit(self):
        with patch(f'{patch_path}.TopicPartition') as TP:
            with patch.object(MockTransaction, 'offset') as offset:
                offset.return_value = 2
                transaction = MockTransaction()
                transaction.commit()
                transaction.producer.send_offsets_to_transaction.assert_called_with([TP(transaction.topic(), transaction.partition(), transaction.offset() + 1)], transaction.consumer.consumer_group_metadata())
                transaction.producer.commit_transaction.assert_called()
                transaction.producer.poll.assert_called_with(0)
                assert transaction._committed
                assert not transaction._active_transaction

    def test_commit__dont_mark_committed(self):
        with patch(f'{patch_path}.TopicPartition') as TP:
            with patch.object(MockTransaction, 'offset') as offset:
                offset.return_value = 2
                transaction = MockTransaction()
                transaction.commit(mark_committed=False)
                transaction.producer.send_offsets_to_transaction.assert_called_with([TP(transaction.topic(), transaction.partition(), transaction.offset() + 1)], transaction.consumer.consumer_group_metadata())
                transaction.producer.commit_transaction.assert_called()
                transaction.producer.poll.assert_called_with(0)
                assert not transaction._committed
                assert transaction._active_transaction

    def test_commit__no_commit(self):
        transaction = MockTransaction()
        transaction._active_transaction = True
        transaction._committed = True
        # just confirm the first function in the block is not called cuz i'm lazy
        transaction.producer.send_offsets_to_transaction.assert_not_called()


class TestGtfoApp:

    def test__get_cluster_name(self, test_topic, cluster_name):
        gtfo_app = MockGtfoApp()
        assert gtfo_app._get_cluster_name([test_topic]) == cluster_name

    def test__get_transactional_producer(self, test_topic, topic_schema_dict, cluster_name):
        schema_reg_obj = MagicMock()
        producer_expected = MagicMock()
        gtfo_app = MockGtfoApp(schema_registry=schema_reg_obj)
        with patch(f'{patch_path}.get_producers') as mock_producer:
            mock_producer.return_value = producer_expected
            producer_actual = gtfo_app._get_transactional_producer(topic_schema_dict, schema_reg_obj, cluster_name)
            producer_actual.init_transactions.assert_called()
            assert producer_expected == producer_actual

    def test__get_transactional_consumer(self, test_topic, cluster_name):
        schema_reg_obj = MagicMock()
        gtfo_app = MockGtfoApp(schema_registry=schema_reg_obj)
        mock_consumer_configs = {'group.id': 'id'}
        mock_kafka_configs = {'bootstrap.servers': 'url'}
        with patch(f'{patch_path}.DeserializingConsumer') as DC:
            consumer = MagicMock()
            DC.side_effect = [consumer]
            consumer_expected = consumer
            with patch(f'{patch_path}.init_transactional_consumer_configs') as consumer_configs:
                consumer_configs.return_value = mock_consumer_configs
                with patch(f'{patch_path}.get_kafka_configs') as kafka_configs:
                    kafka_configs.return_value = (mock_kafka_configs, 'blah')
                    consumer_actual = gtfo_app._get_transactional_consumer([test_topic], schema_reg_obj, cluster_name)
                    kafka_configs.assert_has_calls([call(cluster_name)])
                consumer_configs.assert_has_calls([call([test_topic], schema_reg_obj, mock_kafka_configs, cluster_name, None)])
            DC.assert_has_calls([call(mock_consumer_configs)])
            consumer_actual.subscribe.assert_has_calls([call([test_topic])])
        assert consumer_expected == consumer_actual

    def test__get_transactional_consumer__no_auto_subscribe(self, test_topic, cluster_name):
        schema_reg_obj = MagicMock()
        gtfo_app = MockGtfoApp(schema_registry=schema_reg_obj)
        mock_consumer_configs = {'group.id': 'id'}
        mock_kafka_configs = {'bootstrap.servers': 'url'}
        with patch(f'{patch_path}.DeserializingConsumer') as DC:
            consumer = MagicMock()
            DC.side_effect = [consumer]
            consumer_expected = consumer
            with patch(f'{patch_path}.init_transactional_consumer_configs') as consumer_configs:
                consumer_configs.return_value = mock_consumer_configs
                with patch(f'{patch_path}.get_kafka_configs') as kafka_configs:
                    kafka_configs.return_value = (mock_kafka_configs, 'blah')
                    consumer_actual = gtfo_app._get_transactional_consumer([test_topic], schema_reg_obj, cluster_name, auto_subscribe=False)
                    kafka_configs.assert_has_calls([call(cluster_name)])
                consumer_configs.assert_has_calls([call([test_topic], schema_reg_obj, mock_kafka_configs, cluster_name, None)])
            DC.assert_has_calls([call(mock_consumer_configs)])
            consumer_actual.subscribe.assert_not_called()
        assert consumer_expected == consumer_actual

    def test__get_transactional_consumer__default_schema(self, test_topic, cluster_name, topic_schema_dict):
        schema_reg_obj = MagicMock()
        gtfo_app = MockGtfoApp(schema_registry=schema_reg_obj)
        mock_consumer_configs = {'group.id': 'id'}
        mock_kafka_configs = {'bootstrap.servers': 'url'}
        with patch(f'{patch_path}.DeserializingConsumer') as DC:
            consumer = MagicMock()
            DC.side_effect = [consumer]
            consumer_expected = consumer
            with patch(f'{patch_path}.init_transactional_consumer_configs') as consumer_configs:
                consumer_configs.return_value = mock_consumer_configs
                with patch(f'{patch_path}.get_kafka_configs') as kafka_configs:
                    kafka_configs.return_value = (mock_kafka_configs, 'blah')
                    consumer_actual = gtfo_app._get_transactional_consumer([test_topic], schema_reg_obj, cluster_name, default_schema=topic_schema_dict)
                    kafka_configs.assert_has_calls([call(cluster_name)])
                consumer_configs.assert_has_calls([call([test_topic], schema_reg_obj, mock_kafka_configs, cluster_name, topic_schema_dict)])
            DC.assert_has_calls([call(mock_consumer_configs)])
            consumer_actual.subscribe.assert_has_calls([call([test_topic])])
        assert consumer_expected == consumer_actual

    def test_consume(self):
        gtfo_app = MockGtfoApp()
        with patch.object(gtfo_app, 'transaction_type') as mock_type:
            mock_type.return_value = 'transaction_obj'
            assert gtfo_app.consume("my_arg", kwargs={"mykwargs": "kwarg"}) == 'transaction_obj'
            assert gtfo_app.transaction == 'transaction_obj'
            mock_type.assert_called_with(gtfo_app.producer, gtfo_app.consumer, "my_arg", metrics_manager=gtfo_app.metrics_manager, timeout=10, kwargs={"mykwargs": "kwarg"})

    def test__app_run_loop(self):
        gtfo_app = MockGtfoApp()
        mock_transaction = MockTransaction()
        gtfo_app.transaction = mock_transaction
        with patch.object(mock_transaction, 'commit') as mock_commit:
            with patch.object(gtfo_app, 'consume') as mock_consume:
                gtfo_app._app_run_loop("my_arg", kwargs={"mykwargs": "kwarg"})
                mock_consume.assert_called_with("my_arg", kwargs={"mykwargs": "kwarg"})
                gtfo_app.app_function.assert_called_with(gtfo_app.transaction, *gtfo_app.app_function_arglist)
            mock_commit.assert_called()

    def test__app_run_loop__no_message(self):
        gtfo_app = MockGtfoApp()
        mock_transaction = MockTransaction()
        gtfo_app.transaction = mock_transaction
        with patch.object(mock_transaction, 'commit') as mock_commit:
            with patch.object(gtfo_app, 'consume') as mock_consume:
                mock_consume.side_effect = NoMessageError
                gtfo_app._app_run_loop("my_arg", kwargs={"mykwargs": "kwarg"})
                mock_consume.assert_called_with("my_arg", kwargs={"mykwargs": "kwarg"})
                gtfo_app.app_function.assert_not_called()
            mock_commit.assert_not_called()
        gtfo_app.producer.poll.assert_called()

    def test__app_run_loop__any_nonhandled_error(self):
        gtfo_app = MockGtfoApp()
        mock_transaction = MockTransaction()
        gtfo_app.transaction = mock_transaction
        with patch.object(mock_transaction, 'commit') as mock_commit:
            with patch.object(gtfo_app, 'consume') as mock_consume:
                mock_consume.side_effect = Exception("Uh Oh!")
                with pytest.raises(Exception):
                    gtfo_app._app_run_loop("my_arg", kwargs={"mykwargs": "kwarg"})
                mock_consume.assert_called_with("my_arg", kwargs={"mykwargs": "kwarg"})
                gtfo_app.app_function.assert_not_called()
            mock_commit.assert_not_called()
        gtfo_app.producer.poll.assert_not_called()

    def test_kafka_cleanup(self):
        gtfo_app = MockGtfoApp()
        with patch(f'{patch_path}.shutdown_cleanup') as cleanup:
            gtfo_app.kafka_cleanup()
            cleanup.assert_called_with(consumer=gtfo_app.consumer)

    def test__app_shutdown(self):
        gtfo_app = MockGtfoApp()
        mock_transaction = MockTransaction()
        gtfo_app.transaction = mock_transaction
        with patch.object(gtfo_app, 'kafka_cleanup') as cleanup:
            gtfo_app._app_shutdown()
            gtfo_app.producer.abort_transaction.assert_not_called()
            cleanup.assert_called()

    def test__app_shutdown__abort(self):
        gtfo_app = MockGtfoApp()
        mock_transaction = MockTransaction()
        mock_transaction._active_transaction = True
        gtfo_app.transaction = mock_transaction
        with patch.object(gtfo_app, 'kafka_cleanup') as cleanup:
            gtfo_app._app_shutdown()
            gtfo_app.producer.abort_transaction.assert_called_with(10)
            cleanup.assert_called()

    def test_run__signalraise(self, health_filepath):
        with patch.object(MockGtfoApp, '_app_run_loop') as mock_loop:
            mock_loop.side_effect = SignalRaise(9)
            with patch.object(MockGtfoApp, '_app_shutdown') as mock_shutdown:
                gtfo_app = MockGtfoApp()
                with patch(f'{patch_path}.log_and_raise_error') as mock_log:
                    gtfo_app.run(health_fp=health_filepath)
                    mock_log.assert_not_called()
                    mock_shutdown.assert_called()

    def test_run__exception(self, health_filepath):
        with patch.object(MockGtfoApp, '_app_run_loop') as mock_loop:
            mock_loop.side_effect = Exception
            with patch.object(MockGtfoApp, '_app_shutdown') as mock_shutdown:
                gtfo_app = MockGtfoApp()
                with patch(f'{patch_path}.log_and_raise_error') as mock_log:
                    gtfo_app.run(health_fp=health_filepath)
                    mock_log.assert_called()
                    mock_shutdown.assert_called()


if __name__ == '__main__':
    pytest.main([__file__])



import pytest
from nubium_utils.confluent_utils.gtfo.rocksdb_utils import RDB


class TestRdb:
    @pytest.fixture(scope='session')
    def rdb_instance(self, tmp_path_factory):
        table_path = tmp_path_factory.mktemp('rdb_table')
        return RDB(table_name='test', table_path=table_path)

    def test_rdb_read_write(self, rdb_instance):
        rdb_instance.write('test_key', 'test_value')
        assert rdb_instance.read('test_key') == 'test_value'

    def test_rdb_write_batch(self, rdb_instance):
        rdb_instance.write_batch({'test_key1': 'test_value1', 'test_key2': 'test_value2'})
        assert rdb_instance.read('test_key1') == 'test_value1'
        assert rdb_instance.read('test_key2') == 'test_value2'
        
    def test_rdb_delete(self, rdb_instance):
        rdb_instance.write('delete_test', 'delete_me')
        assert rdb_instance.read('delete_test') == 'delete_me'
        rdb_instance.delete('delete_test')
        assert rdb_instance.read('delete_test') is None

    def test_rdb_close(self, rdb_instance):
        assert len(rdb_instance.rdb.column_families) > 0
        rdb_instance.close()
        assert len(rdb_instance.rdb.column_families) == 0


if __name__ == '__main__':
    pytest.main([__file__])



from sqlitedict import SqliteDict
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from pathlib import Path
from os import makedirs
import json
from copy import deepcopy
import logging
from time import sleep

LOGGER = logging.getLogger(__name__)


class SqliteGtfo:
    def __init__(self, table_name, table_path=None, auto_init=True, max_pending_writes_count=None, min_cache_count=None, max_cache_count=None):
        if not max_pending_writes_count:
            max_pending_writes_count = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_COUNT']) * 5
        if not min_cache_count:
            min_cache_count = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_COUNT']) * 20
        if not max_cache_count:
            max_cache_count = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_COUNT']) * 50
        self._max_pending_writes_count = max_pending_writes_count
        self._min_cache_count = min_cache_count
        self._max_cache_count = max_cache_count
        self.db = None
        self.table_name = table_name
        self.db_cache = {}
        self.write_cache = {}
        self._current_read = (None, '')
        if not table_path:
            table_path = env_vars()['NU_TABLE_PATH']
        makedirs(table_path, exist_ok=True)
        self.full_db_path = Path(f"{table_path}/{table_name}.sqlite").absolute()

        if auto_init:
            self._init_db()
            self.offset = self._get_offset()

    def _get_offset(self):
        try:
            return int(self._read_db('offset'))
        except:
            return 0

    def _init_db(self):
        LOGGER.debug(f'initing table {self.table_name}')
        db = SqliteDict(self.full_db_path.as_posix(), tablename='gtfo', autocommit=False, journal_mode="WAL")
        self.db = db
        sleep(.1)
        self.db['init'] = 'init'
        self.db.commit()  # confirms db is actually fully init-ed by doing this superfluous commit
        LOGGER.info(f'table {self.table_name} initialized')

    def write(self, key, value):
        self.write_cache[key] = value

    def write_batch(self, batch):
        self.write_cache.update(batch)

    def delete(self, key):
        self.write(key, '-DELETED-')

    def set_offset(self, value):
        self.offset = value

    def _commit_and_cleanup_check(self, recovery_multiplier=1):
        LOGGER.debug(f'Current table write cache count for {self.table_name}: {len(self.write_cache)}')
        if len(self.write_cache) >= (self._max_pending_writes_count * recovery_multiplier):
            self.commit()

    def _cleanup_db_cache_check(self):
        if len(self.db_cache) >= self._max_cache_count:
            self.prune_db_cache()

    def commit_and_cleanup_if_ready(self, recovery_multiplier=None):
        LOGGER.debug(f'Checking if table {self.table_name} is ready to commit or prune cache...')
        if not recovery_multiplier:
            recovery_multiplier = 1
        self._commit_and_cleanup_check(recovery_multiplier=recovery_multiplier)
        self._cleanup_db_cache_check()

    def prune_db_cache(self):
        remove_count = min(int(.25 * len(self.db_cache)), len(self.db_cache) - self._min_cache_count)
        LOGGER.info(f'Pruning {remove_count} from {len(self.db_cache)} table {self.table_name} db cached records')
        self.db_cache = {d: self.db_cache[d] for idx, d in enumerate(self.db_cache) if idx > remove_count}

    def commit(self):
        if self.write_cache:
            LOGGER.info(f'committing {len(self.write_cache)} records in table {self.table_name} write cache')
            for key, value in self.write_cache.items():
                if key != '-DELETED-':
                    self.db[key] = json.dumps(value)
                else:
                    del self.db[key]
            self.db['offset'] = str(self.offset)
            self.db.commit()
            self.db_cache.update(self.write_cache)
            self.write_cache = {}

    def _confirm_previous_read_in_cache(self):
        """
        Adds a key to the db cache if it was read but wasn't followed with a write operation
        """
        old_key = self._current_read[0]
        if old_key and (old_key not in self.write_cache) and (old_key not in self.db_cache):
            self.db_cache[old_key] = self._current_read[1]

    def _read_db(self, key):
        LOGGER.debug(f'{key} not in table {self.table_name} cache; querying the DB')
        try:
            return self.db[key]
        except KeyError:
            return None

    def _read_all(self, key):
        """
        Pops the key off the cache so that the keys are always in last-accessed order. This makes for easy slicing later
        to remove stale entries.
        """
        try:
            return key, self.write_cache[key]
        except KeyError:
            pass
        try:
            return key, self.db_cache.pop(key)
        except KeyError:
            pass
        val = self._read_db(key)
        try:
            return key, json.loads(val)
        except (json.JSONDecodeError, TypeError):
            return key, val

    def read(self, key):
        self._confirm_previous_read_in_cache()
        self._current_read = self._read_all(key)
        return deepcopy(self._current_read[1])

    def close(self):
        self.db.close()



import yaml
import os


class EnvironmentTag(yaml.YAMLObject):
    yaml_tag = u'!Environment'

    @classmethod
    def from_yaml(cls, loader, node):
        return os.environ[node.value]


yaml.SafeLoader.add_constructor(EnvironmentTag.yaml_tag, EnvironmentTag.from_yaml)


def load_yaml_fp(fp):
    if fp.endswith('.yaml') or fp.endswith('.yml'):
        with open(fp, 'r') as f:
            data = yaml.safe_load(f)
            if 'data' in data:
                data = yaml.safe_load(data['data']['nubium-topics.yaml'])
    else:
        data = yaml.safe_load(fp)
    return data




from confluent_kafka import TopicPartition
from nubium_utils.custom_exceptions import NoMessageError
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
import logging
from datetime import datetime
from .gtfo_app import GtfoApp, Transaction


LOGGER = logging.getLogger(__name__)


class BatchTransaction(Transaction):
    def __init__(self, producer, consumer, metrics_manager=None, auto_consume=True, timeout=None, message_batch=None,
                 time_elapse_max_seconds=None, consume_max_count=None, parent_app=None):
        self._time_elapse_start = None

        if not time_elapse_max_seconds:
            time_elapse_max_seconds = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_TIME_SECONDS'])
        if not consume_max_count:
            consume_max_count = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_COUNT'])

        self._time_elapse_max_seconds = time_elapse_max_seconds
        self._consume_max_count = consume_max_count
        if self._time_elapse_max_seconds:
            self._time_elapse_start = datetime.now().timestamp()
        self.message_batch_partition_offset_msg = {}

        if not message_batch:
            message_batch = []
        message = message_batch[-1] if message_batch else None

        self.message_batch = message_batch
        super().__init__(producer, consumer, metrics_manager=metrics_manager, auto_consume=False, timeout=timeout,
                         message=message, parent_app=parent_app)
        if self._timeout > self._time_elapse_max_seconds:
            self._timeout = self._time_elapse_max_seconds
        self._auto_consume(self.message_batch, auto_consume)

    def messages(self):
        return self.message_batch

    def _max_consume_time_continue(self):
        if self._time_elapse_max_seconds:
            seconds_elapsed = datetime.now().timestamp() - self._time_elapse_start
            return seconds_elapsed < self._time_elapse_max_seconds
        return True

    def _max_consume_count_continue(self):
        if self._consume_max_count:
            return len(self.message_batch) < self._consume_max_count
        return True

    def _keep_consuming(self):
        return self._max_consume_count_continue() and self._max_consume_time_continue()

    def consume(self):
        """
        Allows you to (additionally) consume more messages on demand via the consume_max_count.
        Usually only necessary with more complex consumption patterns where bulk is swapped between.

        Note the "max count" is the max size you allow self.message_batch to be, which includes all previous consumes
        on this transaction.
        """

        LOGGER.debug(
            f'While messages, consuming up to {self._consume_max_count} messages for up to {self._time_elapse_max_seconds} seconds!')
        while self._keep_consuming():
            try:
                super().consume()
                self.message_batch.append(self.message)
                self.message_batch_partition_offset_msg[self.message.partition()] = self.message
            except NoMessageError:
                if not self.message_batch:
                    raise
                else:
                    break
        LOGGER.info(f'Finished batch consumption; total messages in batch: {len(self.message_batch)}')

    def produce(self, producer_kwargs, headers_passthrough):
        """Since we can consume more than 1 message at a time,
        you also need to specify the headers since it normally operates on the self.message attribute"""
        super().produce(producer_kwargs, headers_passthrough=headers_passthrough)

    def produce_retry(self, exception=None, message=None):
        if message:
            self.message = message
            super().produce_retry(exception=exception)
        else:
            for msg in self.message_batch:
                self.message = msg
                super().produce_retry(exception=exception)

    def produce_failure(self, exception=None, message=None):
        if message:
            self.message = message
            super().produce_failure(exception=exception)
        else:
            for msg in self.message_batch:
                self.message = msg
                super().produce_failure(exception=exception)

    def commit(self, mark_committed=True):
        offsets_to_commit = [TopicPartition(msg.topic(), msg.partition(), msg.offset() + 1) for msg in
                             self.message_batch_partition_offset_msg.values()]
        self._init_transaction()
        if self._active_transaction:
            LOGGER.debug('Committing per partition...')
            if offsets_to_commit:
                self.producer.send_offsets_to_transaction(offsets_to_commit, self.consumer.consumer_group_metadata())
            self.producer.commit_transaction()
            self.producer.poll(0)
            self._committed = mark_committed
            if self._committed:
                self._active_transaction = False
                LOGGER.debug('All partition transactions committed!')


class GtfoBatchApp(GtfoApp):
    def __init__(self, app_function, consume_topics_list, produce_topic_schema_dict=None,
                 transaction_type=BatchTransaction,
                 app_function_arglist=None, metrics_manager=None, schema_registry=None, cluster_name=None,
                 consumer=None, producer=None):
        super().__init__(app_function, consume_topics_list, produce_topic_schema_dict=produce_topic_schema_dict,
                         transaction_type=transaction_type,
                         app_function_arglist=app_function_arglist, metrics_manager=metrics_manager,
                         schema_registry=schema_registry, cluster_name=cluster_name, consumer=consumer,
                         producer=producer)



import pytest
from unittest.mock import patch, MagicMock, call
import os
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.confluent_utils.message_utils import *


class TestMessageUtils:
    def test_handle_no_messages(self):
        with patch('nubium_utils.confluent_utils.message_utils.confirm_produce') as confirm:
            handle_no_messages()
            confirm.assert_not_called()

            producers = MagicMock()
            handle_no_messages(producers=producers)
            confirm.assert_called_with(producers)

    def test_shutdown_cleanup(self):
        with patch('nubium_utils.confluent_utils.message_utils.confirm_produce') as confirm:
            consumer = MagicMock()
            producers = MagicMock()
            shutdown_cleanup(producers=producers, consumer=consumer)

            consumer.close.assert_called()
            confirm.assert_called_with(producers, timeout=30)


if __name__ == '__main__':
    pytest.main([__file__])



import logging
import os.path
from datetime import datetime
from os import environ, path
from shutil import rmtree
from time import sleep

import psutil
import pytest

from nubium_utils.confluent_utils import KafkaToolbox

LOGGER = logging.getLogger(__name__)

eloqua_retriever_timestamp = {
    "name": "EloquaRetrieverTimestamp",
    "type": "record",
    "fields": [
        {"name": "timestamp", "type": "string", "default": ""}
    ]
}

kafka_toolbox = KafkaToolbox()


@pytest.fixture()
def app_wait():
    sleep(10)


@pytest.fixture()
def process_wait():
    sleep(10)


@pytest.fixture()
def delete_app_table():
    if path.exists(environ['NU_TABLE_PATH']):
        rmtree(environ['NU_TABLE_PATH'])


@pytest.fixture()
def initialize_timestamp_topic():
    LOGGER.info("Initializing timestamp topic...")
    timestamp = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
    kafka_toolbox.produce_messages(
        topic=environ['TIMESTAMP_TOPIC'],
        schema=eloqua_retriever_timestamp,
        message_list=[dict(headers={"guid": "N/A", "last_updated_by": "dude"}, key="dude_timestamp", value={"timestamp": timestamp})])
    sleep(30)  # wait for app to consume


@pytest.fixture()
def setup_app(request):
    # optional args, set via @pytest.mark.paramatrize('setup_app', [{'env_overrides': {'var': 'val'}], indirect=True)
    kwargs = {'env_overrides': None}
    kwargs.update(getattr(request, 'param', {}))

    LOGGER.info("Initializing app...")
    parent = kafka_toolbox.run_app(skip_sync="true", runtime_env_overrides=kwargs['env_overrides'])  # skip sync since it happens before running integration
    sleep(10)  # wait for app to launch
    return parent


@pytest.fixture()
def teardown_app(setup_app):
    LOGGER.info("Terminating app...")
    children = psutil.Process(setup_app)
    for child in children.children(recursive=True):
        child.terminate()
    sleep(10)  # wait for app to fully stop



import logging
import json
import mmh3
from nubium_utils import parse_headers
from nubium_utils.yaml_parser import load_yaml_fp
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.custom_exceptions import ProduceHeadersException, ProducerTimeoutFailure
from nubium_utils.metrics import MetricsManager
from .confluent_configs import init_producer_configs, get_kafka_configs
from confluent_kafka import SerializingProducer
from confluent_kafka.schema_registry.avro import AvroSerializer


LOGGER = logging.getLogger(__name__)


def produce_partitioner(producer, producer_kwargs):
    key_hash = mmh3.hash(producer_kwargs['key'])
    partition = key_hash % producer.topic_partitions[producer_kwargs['topic']]
    LOGGER.debug(f'KEY HASH - {key_hash}, PARTITION - {partition}')
    return partition


def producer_serializers(schema_dict, schema_registry):
    return {topic: AvroSerializer(schema_registry, json.dumps(schema)) for topic, schema in schema_dict.items() if
            schema}


def _get_producer(topic_schema_dict, cluster_name, schema_registry, transactional):
    producer = SerializingProducer(
        init_producer_configs(topic_schema_dict, schema_registry, get_kafka_configs(cluster_name)[0], cluster_name, transactional))
    setattr(producer, 'schema_dict', producer_serializers(topic_schema_dict, schema_registry))
    topic_yaml = load_yaml_fp(env_vars()['NU_TOPIC_CONFIGS_YAML'])
    setattr(producer, 'topic_partitions', {topic: int(topic_yaml[topic]['configs']['partitions']) for topic in topic_schema_dict})
    return producer


def get_producers(topic_schema_dict, cluster_name, schema_registry, return_as_singular=True, transactional=False):
    producer_topic_dict = {}
    LOGGER.debug('Setting up Kafka Producer')
    if transactional:
        producer_topic_dict = _get_producer(topic_schema_dict, cluster_name, schema_registry, transactional)
    else:
        for topic, schema in topic_schema_dict.items():
            temp_schema_dict = {topic: schema}
            producer = _get_producer(temp_schema_dict, cluster_name, schema_registry, transactional)
            setattr(producer, 'topic', topic)
            LOGGER.debug(f'Producer configuration for topic {topic} complete.')
            producer_topic_dict[topic] = producer
        if return_as_singular:
            if len(producer_topic_dict) == 1:
                return list(producer_topic_dict.values())[0]
    LOGGER.debug('Producer setup complete.')
    return producer_topic_dict


def produce_message(producer, producer_kwargs: dict, metrics_manager: MetricsManager = None,
                    consumed_msg_headers_passthrough=None):
    """
    Helper for producing a message with confluent_kafka producers; primarily handles headers passthrough and enforcing
    all required headers fields exist before producing.
    You can pass in the previous message headers via 'consumed_msg_headers_passthrough', which extracts the headers.
    Then, if you wish to overwrite any of them, you can provide your own "headers" keyword in 'producer_kwargs'
    which will take that dict and overwrite any matching key in the 'consumed_msg_headers_passthrough' argument.
    You can also provide a None value for a key to remove it entirely.
    :param producer: A confluent_kafka producer
    :type: confluent_kafka.Producer or confluent_kafka.avro.AvroProducer
    :param producer_kwargs: the kwargs to pass into the producer instance
    :type producer_kwargs: dict
    :param metrics_manager: a MetricsManager object instance
    :type metrics_manager: MetricsManager
    :param consumed_msg_headers_passthrough: confluent_kafka Message.headers() from a consumed msg; will add all to new message
    :type consumed_msg_headers_passthrough: list of tuples, or dict
    """
    required_fields = ['guid', 'last_updated_by']

    if 'header' in producer_kwargs:
        raise KeyError('"header" is not the appropriate key for producing message headers; use "headers" instead')

    headers_out = parse_headers(consumed_msg_headers_passthrough)
    headers_out.update(producer_kwargs.get('headers', {}))
    headers_out = {key: value for key, value in headers_out.items() if value is not None}
    missing_required_keys = [key for key in required_fields if key not in headers_out]
    if missing_required_keys:
        raise ProduceHeadersException(f'Message headers are missing the required key(s): {missing_required_keys}')
    producer_kwargs['headers'] = headers_out

    LOGGER.debug('Polling for previous successful produce callbacks before adding additional produces to queue...')
    producer.poll(0)

    # multi-producer management
    if "topic" not in producer_kwargs and hasattr(producer, 'topic'):
        producer_kwargs.update({'topic': producer.topic})
    if "partition" not in producer_kwargs:
        producer_kwargs.update({'partition': produce_partitioner(producer, producer_kwargs)})
    producer._value_serializer = producer.schema_dict[producer_kwargs['topic']]
    if '__changelog' not in producer_kwargs['topic']:  # TODO: add a separate logger for changelog stuff, but for now it just clutters things
        LOGGER.debug(f'Adding message to the produce queue for [topic, partition] [{producer_kwargs["topic"]}, p{producer_kwargs["partition"]}] - key {repr(producer_kwargs["key"])}')
    producer.produce(**producer_kwargs)
    if metrics_manager:
        metrics_manager.inc_messages_produced(1, producer_kwargs['topic'])
    LOGGER.info(f'Message added to the produce queue; GUID {headers_out.get("guid")}')


def confirm_produce(producers, attempts=3, timeout=20):
    """
    Ensure that messages are actually produced by forcing synchronous processing. Must manually check the events queue
    and see if it's truly empty since flushing timeouts do not actually raise an exception for some reason.

    NOTE: Only used for synchronous producing, which is dramatically slower than asychnronous.
    """
    if isinstance(producers, dict):
        producers = producers.values()
    else:
        producers = [producers]
    for producer in producers:
        attempt = 1
        while producer.__len__() > 0:
            if attempt <= attempts:
                LOGGER.debug(f"Produce flush attempt: {attempt} of {attempts}")
                producer.flush(timeout=timeout)
                attempt += 1
            else:
                raise ProducerTimeoutFailure



"""
Container for standardized application monitoring gauges
"""
from prometheus_client import Gauge, CollectorRegistry
from nubium_utils.general_runtime_vars import env_vars
from nubium_utils.metrics.metrics_pusher import MetricsPusher


class Metric:
    def __init__(self, metric_name, registry, description='', job=None, app=None, additional_labels=None):
        if not job:
            job = env_vars()['NU_HOSTNAME']
        if not app:
            app = env_vars()['NU_APP_NAME']
        if not additional_labels:
            additional_labels = []

        self.name = metric_name
        self.app = app
        self.job = job
        self.additional_labels = list(sorted(additional_labels))
        self.all_labels = ['app', 'job'] + self.additional_labels
        self.store = Gauge(self.name, description, labelnames=self.all_labels, registry=registry)

    def _labels(self, label_dict):
        if not self.additional_labels:
            label_dict = {}
        return self.store.labels(self.app, self.job, *[i[1] for i in sorted(label_dict.items())])

    def inc(self, number=1, label_dict=None):
        self._labels(label_dict).inc(number)

    def set(self, number, label_dict=None):
        self._labels(label_dict).set(number)


class MetricsManager:
    """
    Coordinates Prometheus monitoring for a Kafka client application

    Creates and manages Metric instances and pushes their metrics
    """

    def __init__(self, registry=None, metrics_pusher: MetricsPusher = None):
        """
        Initializes monitor and Metric classes
        """
        if not metrics_pusher:
            metrics_pusher = MetricsPusher()
        if not registry:
            registry = CollectorRegistry()

        self.metrics_pusher = metrics_pusher
        self.registry = registry
        self._metrics = {}

        self.new_metric('messages_consumed', description='Messages consumed since application start', additional_labels=['topic']),
        self.new_metric('messages_produced', description='Messages produced since application start', additional_labels=['topic']),
        self.new_metric('message_errors', description='Exceptions caught when processing messages', additional_labels=['exception']),
        self.new_metric('external_requests', description='Network calls to external services', additional_labels=['request_to', 'request_endpoint', 'request_type', 'is_bulk', 'status_code']),
        self.new_metric('seconds_behind', description='Elapsed time since the consumed message was originally produced')

    def __getattr__(self, name):
        try:
            return self._metrics[name]
        except:
            return super().__getattribute__(name)

    @property
    def metric_names(self):
        return [metric.name for metric in self._metrics.values()]

    def new_metric(self, metric_name, description='', additional_labels=None):
        self._metrics[metric_name] = Metric(metric_name, self.registry, description=description, additional_labels=additional_labels)

    def inc_metric(self, metric_name, number=1, label_dict=None):
        self._metrics[metric_name].inc(number=number, label_dict=label_dict)

    def set_metric(self, metric_name, number, label_dict=None):
        self._metrics[metric_name].set(number, label_dict=label_dict)

    def push_metrics(self):
        self.metrics_pusher.set_metrics_pod_ips()
        self.metrics_pusher.push_metrics(self.registry)

    # ------------------------------- BELOW: BACKWARDS COMPATABILITY; remove in v4.0? --------------------------------

    def register_custom_metric(self, name, description):
        """ Register a custom metric to be used with inc_custom_metric later. """
        self.new_metric(name, description=description)

    def inc_custom_metric(self, name, amount=1):
        """
        Increases a registered custom metric by the amount specified (defaults to 1).

        `name`'s value must be registered in advance by calling register_custom_metric.
        """
        if name not in self._metrics:
            raise ValueError(f"'{name}' was not registered via register_custom_metric.")
        self.inc_metric(name, number=amount)

    def inc_messages_consumed(self, number_of_messages, topic):
        """
        Increases the messages_consumed gauge with default labels
        """
        self.inc_metric('messages_consumed', number=number_of_messages, label_dict={'topic': topic})

    def inc_messages_produced(self, number_of_messages, topic):
        """
        Increases the messages_consumed gauge with default labels
        """
        self.inc_metric('messages_produced', number=number_of_messages, label_dict={'topic': topic})

    def inc_message_errors(self, exception):
        """
        Increases the error gauge with default label and label of the exception
        """
        self.inc_metric('message_errors', label_dict={'exception': exception.__class__.__name__})

    def inc_external_requests(self, request_to=None, request_endpoint=None, request_type=None, is_bulk=0, status_code=200):
        """
        Increases the external requests gauge.
        """
        self.inc_metric(
            'external_requests',
            label_dict={
                "request_to": request_to,
                "request_endpoint": request_endpoint,
                "request_type": request_type,
                "is_bulk": is_bulk,
                "status_code": status_code})

    def set_seconds_behind(self, seconds_behind):
        """
        Sets the seconds_behind gauge with default labels
        """
        self.set_metric('seconds_behind', seconds_behind)



import pytest
from unittest.mock import patch, MagicMock, call
import os
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.confluent_utils.confluent_configs import *

patch_path = 'nubium_utils.confluent_utils.confluent_configs'


def required_env_vars():
    return {
        "NU_LOGLEVEL": "DEBUG",
        "NU_HOSTNAME": "test-app-0",
        "NU_APP_NAME": "test_app",
        "NU_MP_PROJECT": "none",
        "NU_MP_CLUSTER": "none",
        "NU_DO_METRICS_PUSHING": "false",
        "NU_SCHEMA_REGISTRY_URL": "sr_url",
        "NU_SCHEMA_REGISTRY_USERNAME": "sr_un",
        "NU_SCHEMA_REGISTRY_PASSWORD": "sr_pw",
        "NU_TOPIC_CONFIGS_YAML": '{"test_topic_0": {"cluster": "cluster_0"}, "test_topic_1": {"cluster": "cluster_1"}}',
        "NU_KAFKA_CLUSTERS_CONFIGS_YAML": '{"cluster_0": {"url": "www", "username": "un", "password": "pw"}}',
    }


@pytest.fixture
def set_env_vars(request):
    """
    Allows you to set additional env vars per test, but requires an empty dict if you dont
    """
    evars = required_env_vars()
    evars.update(request.param)
    env_patch = patch.dict('os.environ', evars)
    env_patch.start()
    env_vars._reload()


class TestConfluentConfigs:
    # @pytest.mark.parametrize('patched_env_vars', [{'NU_KAFKA_CLUSTERS_CONFIGS_YAML': '{"cluster_0": {"url": "www", "username": "un", "password": "pw"}}'}], indirect=True)
    @pytest.mark.parametrize('set_env_vars', [{}], indirect=True)
    def test_init_sasl_configs(self, set_env_vars):
        assert init_sasl_configs('cluster_0') == {
            "security.protocol": "sasl_ssl",
            "sasl.mechanisms": "PLAIN",
            "sasl.username": 'un',
            "sasl.password": 'pw'
        }

    @pytest.mark.parametrize('set_env_vars', [{}], indirect=True)
    def test_init_schema_registry_configs__auth(self, set_env_vars):
        assert init_schema_registry_configs() == {"schema.registry.url": "https://sr_un:sr_pw@sr_url"}

        with patch('nubium_utils.confluent_utils.confluent_configs.SchemaRegistryClient') as SRC:
            init_schema_registry_configs(as_registry_object=True)
            SRC.assert_called_with({"url": "https://sr_un:sr_pw@sr_url"})

    @pytest.mark.parametrize('set_env_vars', [{'NU_SCHEMA_REGISTRY_URL': 'localhost', 'NU_SCHEMA_REGISTRY_USERNAME': ''}], indirect=True)
    def test_init_schema_registry_configs__localhost(self, set_env_vars):
        assert init_schema_registry_configs() == {"schema.registry.url": "http://localhost"}
        with patch('nubium_utils.confluent_utils.confluent_configs.SchemaRegistryClient') as SRC:
            init_schema_registry_configs(as_registry_object=True)
            SRC.assert_called_with({"url": "http://localhost"})
    
    @pytest.mark.parametrize('set_env_vars', [{}], indirect=True)
    def test_init_producer_configs(self, set_env_vars):
        schema_reg_obj = MagicMock()
        sasl_config = {'sasl': 'config'}
        with patch(f'{patch_path}.AvroSerializer') as AS:
            expected = {
                **sasl_config,
                "bootstrap.servers": "www",
                "on_delivery": produce_message_callback,
                "partitioner": "murmur2_random",
                "key.serializer": AS(schema_reg_obj, schema_str='{"type": "string"}'),
                "value.serializer": throwaway,
                "enable.idempotence": "true",
                "acks": "all"
            }
            
            assert init_producer_configs({'test_topic_0': 'schema'}, schema_reg_obj, sasl_config, 'cluster_0', False) == expected
            AS.assert_called_with(schema_reg_obj, schema_str='{"type": "string"}')

            # no cluster provided
            assert init_producer_configs({'test_topic_0': 'schema'}, schema_reg_obj, sasl_config, transactional=False) == expected
            AS.assert_called_with(schema_reg_obj, schema_str='{"type": "string"}')

    @pytest.mark.parametrize('set_env_vars', [{}], indirect=True)
    def test_init_producer_configs__transactional(self, set_env_vars):
        schema_reg_obj = MagicMock()
        sasl_config = {'sasl': 'config'}
        with patch(f'{patch_path}.AvroSerializer') as AS:
            expected = {
                **sasl_config,
                "bootstrap.servers": "www",
                "on_delivery": produce_message_callback,
                "partitioner": "murmur2_random",
                "key.serializer": AS(schema_reg_obj, schema_str='{"type": "string"}'),
                "value.serializer": throwaway,
                "enable.idempotence": "true",
                "transaction.timeout.ms": 120000,
                "transactional.id": "test-app-0",
                "acks": "all"}

            assert init_producer_configs({'test_topic_0': 'schema'}, schema_reg_obj, sasl_config, 'cluster_0', True) == expected
            AS.assert_called_with(schema_reg_obj, schema_str='{"type": "string"}')
            
    @pytest.mark.parametrize('set_env_vars', [{}], indirect=True)
    def test_init_producer_configs__bad_topic_dict(self, set_env_vars):
        schema_reg_obj = MagicMock()
        sasl_config = {'sasl': 'config'}
        with patch(f'{patch_path}.AvroSerializer') as AS:
            with pytest.raises(ValueError):
                init_producer_configs({'test_topic_0': 'schema', 'test_topic_1': 'schema'}, schema_reg_obj, sasl_config)

    @pytest.mark.parametrize('set_env_vars', [{}], indirect=True)
    def test_init_transactional_consumer_configs(self, set_env_vars):
        schema_reg_obj = MagicMock()
        sasl_config = {'sasl': 'config'}

        with patch(f'{patch_path}.AvroDeserializer') as AD:
            expected = {
                **sasl_config,
                "bootstrap.servers": "www",
                "group.id": "test_app",
                "on_commit": consume_message_callback,
                "auto.offset.reset": 'latest',
                "isolation.level": "read_committed",
                "enable.auto.commit": False,
                "enable.auto.offset.store": False,
                "key.deserializer": AD(schema_reg_obj, schema_str='{"type": "string"}'),
                "value.deserializer": AD(schema_reg_obj, schema_str='{"schema": "schema"}'),
                "partition.assignment.strategy": 'cooperative-sticky',
                "max.poll.interval.ms": 60000 * (2+0),
                "session.timeout.ms": 1000 * 90,
                "message.max.bytes": 2 * (2 ** 20),
                "fetch.max.bytes": 3 * (2 ** 20),
                "queued.max.messages.kbytes": 10 * (2 ** 10),
            }

            assert init_transactional_consumer_configs(['test_topic_0'], schema_reg_obj, sasl_config, 'cluster_0', {"schema": "schema"}) == expected
            AD.assert_has_calls([call(schema_reg_obj, schema_str='{"type": "string"}'), call(schema_reg_obj, schema_str='{"schema": "schema"}')])

            # no cluster provided
            assert init_transactional_consumer_configs(['test_topic_0'], schema_reg_obj, sasl_config, schema={"schema": "schema"}) == expected
            AD.assert_has_calls([call(schema_reg_obj, schema_str='{"type": "string"}'), call(schema_reg_obj, schema_str='{"schema": "schema"}')])

    @pytest.mark.parametrize('set_env_vars', [{}], indirect=True)
    def test_init_transactional_consumer_configs__bad_topic_list(self, set_env_vars):
        schema_reg_obj = MagicMock()
        sasl_config = {'sasl': 'config'}
        with patch(f'{patch_path}.AvroDeserializer') as AD:
            with pytest.raises(ValueError):
                init_transactional_consumer_configs(['test_topic_0', 'test_topic_1'], schema_reg_obj, sasl_config)


if __name__ == '__main__':
    pytest.main([__file__])



import logging
from nubium_utils.metrics import start_pushing_metrics
from .confluent_runtime_vars import env_vars
from nubium_utils.yaml_parser import load_yaml_fp
from nubium_utils.custom_exceptions import WrappedSignals
from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import AvroSerializer, AvroDeserializer
import json


# fixes a bug in confluent-kafka TODO: keep a look out for this in >1.8.2, should be fixed since I took it from a MR.
def patched_schema_loads(schema_str):
    from confluent_kafka.schema_registry.avro import Schema
    schema_str = schema_str.strip()
    if schema_str[0] != "{" and schema_str[0] != "[":
        schema_str = '{"type":' + schema_str + '}'
    return Schema(schema_str, schema_type='AVRO')
import confluent_kafka
confluent_kafka.schema_registry.avro._schema_loads = patched_schema_loads


LOGGER = logging.getLogger(__name__)
wrapped_signals = WrappedSignals()  # WrappedSignals needs to be initialized (somewhere) to catch signals


def init_sasl_configs(cluster_name):
    cluster_configs = load_yaml_fp(env_vars()['NU_KAFKA_CLUSTERS_CONFIGS_YAML']).get(cluster_name, {})
    if cluster_configs.get('username'):
        LOGGER.info(f'Using SASL-authenticated producers/consumers for {cluster_name}!')
        return {
            "security.protocol": "sasl_ssl",
            "sasl.mechanisms": "PLAIN",
            "sasl.username": cluster_configs['username'],
            "sasl.password": cluster_configs['password']
        }
    else:
        LOGGER.info(f'NO Authentication/encryption for producers/consumers in {cluster_name}!')
    return {}


def init_schema_registry_configs(as_registry_object=True):
    """
    Provides the avro schema config
    :return: dict, None
    """
    config = None
    if as_registry_object:
        url = 'url'
    else:
        url = 'schema.registry.url'
    LOGGER.info(f'Using the schema server at {env_vars()["NU_SCHEMA_REGISTRY_URL"]}')
    # RHOSAK registry
    if env_vars()['NU_SCHEMA_REGISTRY_USERNAME']:
        config = {url: f"https://{env_vars()['NU_SCHEMA_REGISTRY_SERVER']}"}
    else:  # local
        config = {url: f"http://{env_vars()['NU_SCHEMA_REGISTRY_SERVER']}"}

    if as_registry_object:
        return SchemaRegistryClient(config)
    else:
        return config
    

def throwaway():
    return None


def produce_message_callback(error, message):
    """
    Logs the returned message from the Broker after producing
    NOTE: Headers not supported on the message for callback for some reason.
    NOTE: Callback requires these args
    """
    if error:
        LOGGER.critical(error)
    else:
        LOGGER.debug('Producer Callback - Message produced successfully')


def init_producer_configs(topic_schema_dict, schema_registry, sasl_configs=None, cluster_name=None, transactional=False):
    """
    Sets up a single producer.
    """
    if topic_schema_dict and not cluster_name:
        topic_cluster_map = load_yaml_fp(env_vars()['NU_TOPIC_CONFIGS_YAML'])
        check_topics = set([topic for topic in topic_schema_dict if topic in topic_cluster_map])
        unmapped = set(topic_schema_dict.keys()) - check_topics
        if unmapped:
            LOGGER.warning(f'The topic(s) {unmapped} in the provided producer schema dict mapping are not included in'
                           f' the cluster topic map. There are valid use cases for this, such as dynamically adding'
                           f' new topics at runtime with a predetermined schema, or even just testing a new topic on'
                           f' the fly. Either way, make sure this is not a mistake!')
        cluster_name = set([topic_cluster_map[topic]['cluster'] for topic in check_topics])
        if len(cluster_name) == 1:
            cluster_name = list(cluster_name)[0]
        else:
            raise ValueError('Cluster count for provided producer topics did not equal 1. '
                             f'These topics must be on the same cluster to init their configs. Topics:\n{check_topics}.')
    if not sasl_configs:
        sasl_configs = {}
    config = {
        **sasl_configs,
        "bootstrap.servers": load_yaml_fp(env_vars()['NU_KAFKA_CLUSTERS_CONFIGS_YAML'])[cluster_name]['url'],
        "on_delivery": produce_message_callback,
        "partitioner": "murmur2_random",
        "key.serializer": AvroSerializer(schema_registry, schema_str='{"type": "string"}'),
        "value.serializer": throwaway,
        "enable.idempotence": "true",
        "acks": "all"}
    if transactional: 
        config["transactional.id"] = env_vars()['NU_HOSTNAME']
        config["transaction.timeout.ms"] = (int(env_vars()['NU_CONSUMER_TIMEOUT_LIMIT_MINUTES']) + int(env_vars()['NU_CONSUMER_TIMESTAMP_OFFSET_MINUTES'])) * 60000
    LOGGER.debug(f"Producer config: {json.dumps({k:v for k,v in config.items() if 'sasl' not in k and isinstance(v, str)}, indent=4)}")
    return config


def consume_message_callback(error, partitions):
    """
    Logs the info returned when a successful commit is performed
    NOTE: Callback requires these args
    """
    if error:
        LOGGER.critical(error)
    else:
        LOGGER.debug('Consumer Callback - Message consumption committed successfully')


def init_transactional_consumer_configs(topics, schema_registry, sasl_configs=None, cluster_name=None, schema=None):
    """
    Assumes topics are all in the same cluster.
    """
    if not cluster_name:
        topic_configs = load_yaml_fp(env_vars()['NU_TOPIC_CONFIGS_YAML'])
        cluster_name = set([topic_configs[topic]['cluster'] for topic in topics])
        if len(cluster_name) == 1:
            cluster_name = list(cluster_name)[0]
        else:
            raise ValueError('Cluster count for provided consumer topics did not equal 1. '
                             f'Please confirm your topics are on the same cluster. Topics:\n{topics}')
    if not sasl_configs:
        sasl_configs = {}
    max_time_between_consumes_mins = int(env_vars()['NU_CONSUMER_TIMEOUT_LIMIT_MINUTES']) + int(env_vars()['NU_CONSUMER_TIMESTAMP_OFFSET_MINUTES'])
    session_timeout = int(env_vars()['NU_CONSUMER_HEARTBEAT_TIMEOUT_SECONDS']) * 1000
    schema = json.dumps(schema) if schema else None
    config = {
        # Connection Settings
        **sasl_configs,
        "bootstrap.servers": load_yaml_fp(env_vars()['NU_KAFKA_CLUSTERS_CONFIGS_YAML'])[cluster_name]['url'],

        # General Consumer Settings
        "group.id": env_vars()['NU_APP_NAME'],
        "on_commit": consume_message_callback,
        "auto.offset.reset": env_vars()['NU_CONSUMER_AUTO_OFFSET_RESET'],

        # Transaction Settings
        "isolation.level": "read_committed",
        "enable.auto.commit": False,
        "enable.auto.offset.store": False,

        # Registry Serialization Settings
        "key.deserializer": AvroDeserializer(schema_registry, schema_str='{"type": "string"}'),
        "value.deserializer": AvroDeserializer(schema_registry, schema_str=schema),

        # Performance Settings
        "partition.assignment.strategy": 'cooperative-sticky',
        "max.poll.interval.ms": 60000 * max_time_between_consumes_mins,  # Max time between poll() calls before considered dead.
        "session.timeout.ms": session_timeout,  # need at least 1 heartbeat within a session to be considered alive.
        # "heartbeat.interval.ms": (session_timeout // 3) - 1000,  # recommended so that 3 failed heartbeats == bad consumer.
        "message.max.bytes": int(env_vars()['NU_CONSUMER_MESSAGE_BATCH_MAX_MB']),
        "fetch.max.bytes": int(env_vars()['NU_CONSUMER_MESSAGE_TOTAL_MAX_MB']),
        "queued.max.messages.kbytes": int(env_vars()['NU_CONSUMER_MESSAGE_QUEUE_MAX_MB']),
    }
    LOGGER.debug(f"Consumer config: {json.dumps({k:str(v) for k,v in config.items() if 'sasl' not in k}, indent=4)}")
    return config


def init_metrics_pushing(metrics_manager):
    if env_vars()['NU_DO_METRICS_PUSHING'].lower() == 'true':
        LOGGER.info('Metric pushing to gateway ENABLED')
        start_pushing_metrics(metrics_manager, int(env_vars()['NU_METRICS_PUSH_RATE']))
    else:
        LOGGER.info('Metric pushing to gateway DISABLED')


def get_kafka_configs(cluster_name):
    return init_sasl_configs(cluster_name), init_schema_registry_configs(as_registry_object=False)



class MockMessage:
    def __init__(self, key=None, value=None, headers=None, topic=None, timestamp=None, partition=None, offset=None, error=None):
        if not key:
            key = 'test_key'
        self.k = key
        if not value:
            value = 'test_value'
        self.v = value
        if not headers:
            headers = {'test_header_key': 'test_header_value'}
        self.h = headers
        if not topic:
            topic = 'test_topic'
        self.t = topic
        if not timestamp:
            timestamp = ('timestamp', 1642741200000) # Jan 21 2022 00:00:00.000 UTC
        self.time = timestamp
        if not partition:
            partition = 0
        self.p = partition
        if not offset:
            offset = 1
        self.of = offset
        self.e = error
    
    def key(self):
        return self.k

    def value(self):
        return self.v

    def headers(self):
        return self.h

    def topic(self):
        return self.t

    def timestamp(self):
        return self.time

    def partition(self):
        return self.p

    def offset(self):
        return self.of

    def error(self):
        return self.e



from .general_utils import generate_guid, parse_headers, log_and_raise_error
from .metrics import MetricsPusher, MetricsManager
from .logging_utils import init_logger
from .dict_manipulator import DictManipulator

init_logger(__name__)


import json
from confluent_kafka import TopicPartition, DeserializingConsumer, KafkaException
from nubium_utils.general_utils import parse_headers, log_and_raise_error
from nubium_utils.custom_exceptions import NoMessageError, SignalRaise, RetryTopicSend, FailureTopicSend, MaxRetriesReached
from nubium_utils.metrics import MetricsManager
from nubium_utils.yaml_parser import load_yaml_fp
from nubium_utils.confluent_utils.consumer_utils import consume_message
from nubium_utils.confluent_utils.producer_utils import produce_message, get_producers
from nubium_utils.confluent_utils.message_utils import shutdown_cleanup
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.confluent_utils.confluent_configs import init_transactional_consumer_configs, init_schema_registry_configs, get_kafka_configs, init_metrics_pushing
import logging
from copy import deepcopy
from os import remove
from wrapt_timeout_decorator import timeout as wrapt_timeout

LOGGER = logging.getLogger(__name__)


class GracefulTransactionFailure(Exception):
    def __init__(self):
        pass


class FatalTransactionFailure(Exception):
    def __init__(self):
        pass


def handle_kafka_exception(kafka_error):
    LOGGER.error(kafka_error)
    if kafka_error.args[0].code == 'ILLEGAL GENERATION':
        LOGGER.info('The consumer group generation id is invalid (likely due to a rebalance call), aborting transaction')
        raise FatalTransactionFailure
    retriable = kafka_error.args[0].retriable()
    abort = kafka_error.args[0].txn_requires_abort()
    # TODO: take advantage of retriable
    LOGGER.info(f'KafkaException: is retriable? - {retriable}, should abort? - {abort}')
    if retriable or abort:
        raise GracefulTransactionFailure
    else:
        raise FatalTransactionFailure


class Transaction:
    def __init__(self, producer, consumer, metrics_manager=None, auto_consume=True, timeout=None, message=None, parent_app=None):
        if not timeout:
            timeout = int(env_vars()['NU_CONSUMER_POLL_TIMEOUT'])
        self.producer = producer
        self.consumer = consumer
        self.message = message
        self.metrics_manager = metrics_manager
        self._timeout = timeout
        self._allow_auto_consume = auto_consume

        # this optional attribute should never be referenced/used by class methods here to keep transactions independent of the app
        # only intended for runtime access of the app instance by the transaction
        self.app = parent_app

        self._retry_offsets = {}
        self._committed = False  # mostly used for signaling to GtfoApp that there's currently a transaction and handling automatic committing
        self._active_transaction = False
        self._auto_consume(self.message, self._allow_auto_consume)

    def _refresh_transaction_fields(self):
        self._retry_offsets = {}
        self._committed = False  # mostly used for signaling to GtfoApp that there's currently a transaction and handling automatic committing
        self._active_transaction = False
        self._auto_consume(self.message, self._allow_auto_consume)

    def _auto_consume(self, has_input, should_consume):
        """ Consume message if initialized without one (and allowed to) """
        if not has_input and should_consume:
            self.consume()

    def messages(self):  # here for a consistent way to access message(s) currently being managed for when you subclass
        """ For a standardized way to access message(s) consumed pertaining to this transaction """
        return self.message

    def _get_consumer_partition_assignment(self):
        assignments = self.consumer.assignment()
        assignments = {topic: [int(obj.partition) for obj in assignments if obj.topic == topic] for topic in set([obj.topic for obj in assignments])}
        return assignments

    def _rollback_transaction_consumption(self):
        LOGGER.info('Rolling back consumer state...')
        assignments = self._get_consumer_partition_assignment()
        for topic, partitions in self._retry_offsets.items():
            for partition, offset in partitions.items():
                if partition in assignments.get(topic, []):
                    LOGGER.info(f"Reversing topic {topic} partition {partition} back to offset {offset}")
                    self.consumer.seek(TopicPartition(topic=topic, partition=partition, offset=offset))
                    LOGGER.info(f"Consumer set topic {topic} partition {partition} to offset {self.consumer.position([TopicPartition(topic=topic, partition=partition)])[0].offset}")

    def mark_retry_offset(self):
        if self.topic() not in self._retry_offsets:
            self._retry_offsets[self.topic()] = {}
        if self.partition() not in self._retry_offsets[self.topic()]:
            self._retry_offsets[self.topic()][self.partition()] = self.offset()

    def consume(self):
        self.message = consume_message(self.consumer, self.metrics_manager, self._timeout)
        self.mark_retry_offset()

    def key(self):
        return deepcopy(self.message.key())

    def value(self):
        return deepcopy(self.message.value())

    def headers(self):
        return deepcopy(parse_headers(self.message.headers()))

    def topic(self):
        return self.message.topic()

    def partition(self):
        return self.message.partition()

    def offset(self):
        return self.message.offset()

    @wrapt_timeout(15)
    def _abort_transaction(self):
        self.producer.abort_transaction(10)

    def abort_active_transaction(self):
        LOGGER.debug('Aborting any open transactions, if needed.')
        reset_producer = False
        if self._active_transaction:
            try:
                LOGGER.info('Aborting transaction.')
                self._abort_transaction()
                reset_producer = True
                LOGGER.info('Abortion complete!')
                self.producer.poll(0)
                self._rollback_transaction_consumption()
                self._refresh_transaction_fields()
            except KafkaException as kafka_error:
                LOGGER.info(f"Failed to abort transaction: {kafka_error}")
                pass
        return reset_producer

    def _init_transaction(self):
        """ Mark that a transaction is now underway. Triggered by trying to produce or commit a message """
        try:
            if not self._active_transaction and not self._committed:
                LOGGER.info('Beginning Transaction...')
                self.producer.begin_transaction()
                self._active_transaction = True
                self.producer.poll(0)
        except KafkaException as kafka_error:
            handle_kafka_exception(kafka_error)

    def produce(self, producer_kwargs, headers_passthrough=None):
        if not headers_passthrough:
            headers_passthrough = self.headers()
        self._init_transaction()
        produce_message(self.producer, producer_kwargs, self.metrics_manager, headers_passthrough)
        self.producer.poll(0)

    def produce_retry(self, exception=None):
        retry_topic = None
        headers = self.headers()
        guid = headers['guid']
        kafka_retry_count = int(headers.get('kafka_retry_count', '0'))

        if kafka_retry_count < int(env_vars()['NU_RETRY_COUNT_MAX']):
            headers['kafka_retry_count'] = str(kafka_retry_count + 1)
            retry_topic = env_vars()['NU_CONSUME_TOPICS']
        else:
            headers['kafka_retry_count'] = '0'
            retry_topic = env_vars().get('NU_PRODUCE_RETRY_TOPICS', '')

        if retry_topic:
            if not exception:
                exception = RetryTopicSend()
            LOGGER.warning('; '.join([str(exception), f'retrying GUID {guid}']))
            self.produce(dict(
                topic=retry_topic,
                value=self.value(),
                key=self.key(),
                headers=headers))
        else:
            if not exception:
                exception = FailureTopicSend()
            LOGGER.error('; '.join([str(exception), f'GUID {guid}']))
            self.produce_failure(exception=MaxRetriesReached())

    def produce_failure(self, exception=None):
        headers = self.headers()
        guid = headers['guid']
        headers['kafka_retry_count'] = '0'
        failure_topic = env_vars()['NU_PRODUCE_FAILURE_TOPICS']

        if not exception:
            exception = FailureTopicSend()
        LOGGER.error('; '.join([type(exception).__name__, str(exception), f'failing GUID {guid}']))
        headers["exception"] = json.dumps({"name": type(exception).__name__, "description": str(exception)})

        LOGGER.debug(f'Adding a message to the produce queue for deadletter/failure topic {env_vars()["NU_PRODUCE_FAILURE_TOPICS"]}')
        self.produce(dict(
            topic=failure_topic,
            value=self.value(),
            key=self.key(),
            headers=headers))
        LOGGER.info(f'Message added to the deadletter/failure topic produce queue; GUID {guid}')

    def commit(self, mark_committed=True):
        """ Allows manual commits (safety measures in place so that you cant commit the same message twice)."""
        LOGGER.info('Committing Transaction...')
        self._init_transaction()
        try:
            if self.message:
                # NOTE - if consuming from a transactional topic, the offset+1 commit will lead to off-by-one error
                # which makes it look like it has lag, but it doesn't.
                self.producer.send_offsets_to_transaction(
                    [TopicPartition(self.topic(), self.partition(), self.offset() + 1)],
                    self.consumer.consumer_group_metadata(), 8)
            if self._active_transaction:
                self.producer.commit_transaction(15)
                self.producer.poll(0)
                LOGGER.info('Transaction Committed!')
                self._committed = mark_committed
                if self._committed:
                    self._active_transaction = False
                self._retry_offsets = {}
        except KafkaException as kafka_error:
            handle_kafka_exception(kafka_error)


class GtfoApp:
    """ The main class to use for most GTFO apps. See README for initialization/usage details. """
    def __init__(self, app_function, consume_topics_list, produce_topic_schema_dict=None, transaction_type=Transaction,
                 app_function_arglist=None, metrics_manager=None, schema_registry=None, cluster_name=None, consumer=None, producer=None):
        self.transaction = None
        self._shutdown = False
        if not app_function_arglist:
            app_function_arglist = []
        if not metrics_manager:
            metrics_manager = MetricsManager()
        init_metrics_pushing(metrics_manager)
        if isinstance(consume_topics_list, str):
            consume_topics_list = consume_topics_list.split(',')
        if not produce_topic_schema_dict:  # for when the app is consume-only
            produce_topic_schema_dict = {topic: None for topic in consume_topics_list}
        if not schema_registry:
            schema_registry = init_schema_registry_configs()
        if not cluster_name:
            topic_list = consume_topics_list if consume_topics_list else list(produce_topic_schema_dict.keys())
            cluster_name = self._get_cluster_name(topic_list)
        if not consumer:
            consumer = self._get_transactional_consumer(consume_topics_list, schema_registry, cluster_name)
        if not producer:
            producer = self._get_transactional_producer(produce_topic_schema_dict, schema_registry, cluster_name)

        self.transaction_type = transaction_type
        self.app_function = app_function
        self.app_function_arglist = app_function_arglist
        self.metrics_manager = metrics_manager
        self.produce_topic_schema_dict = produce_topic_schema_dict
        self.schema_registry = schema_registry
        self.cluster_name = cluster_name
        self.consumer = consumer
        self.producer = producer

        self.consume(auto_consume=False)

    def _get_cluster_name(self, consume_topics_list):
        topic = consume_topics_list[0] if isinstance(consume_topics_list, list) else consume_topics_list
        return load_yaml_fp(env_vars()['NU_TOPIC_CONFIGS_YAML'])[topic]['cluster']

    def _get_transactional_producer(self, topic_schema_dict, schema_registry, cluster_name):
        LOGGER.debug('Setting up Kafka Transactional Producer')
        producer = get_producers(topic_schema_dict, cluster_name, schema_registry, transactional=True)
        producer.init_transactions()
        LOGGER.debug('Producer setup complete.')
        return producer

    def reset_gtfo_producer(self):
        self.producer = self._get_transactional_producer(self.produce_topic_schema_dict, self.schema_registry, self.cluster_name)
        self.transaction.producer = self.producer

    def _abort_active_transaction(self):
        if self.transaction.abort_active_transaction():
            self.reset_gtfo_producer()

    def _get_transactional_consumer(self, topics, schema_registry, cluster_name, default_schema=None, auto_subscribe=True):
        LOGGER.debug('Setting up Kafka Transactional Consumer')
        consumer = DeserializingConsumer(
            init_transactional_consumer_configs(topics, schema_registry, get_kafka_configs(cluster_name)[0], cluster_name, default_schema))
        if auto_subscribe:
            consumer.subscribe(topics)  # in case multiple topics are read from
            LOGGER.info(f'Transactional consumer subscribed to topics:\n{topics}')
        return consumer

    def commit(self):
        if not self.transaction._committed:
            LOGGER.debug('Initializing commit...')
            # TODO: Remove all but commit once confirmed that no funny business is happening
            assignments = self.transaction._get_consumer_partition_assignment().get(self.transaction.topic(), [])
            if self.transaction.partition() in assignments:
                self.transaction.commit()
            else:
                LOGGER.error(f'WTF??? msg partition {self.transaction.partition()} not in assignments: {assignments}')
                LOGGER.error('Skipping this message')
                self.transaction.abort_active_transaction()
                self.transaction.message = None

    def consume(self, *args, timeout=10, **kwargs):
        """
        Accepts *args and **kwargs to make it easy to alter functionality of this function in subclasses.
        Public method so that you can manually consume messages if you desire; helpful for debugging.
        """
        self.transaction = self.transaction_type(self.producer, self.consumer, *args, metrics_manager=self.metrics_manager, timeout=timeout, parent_app=self, **kwargs)
        return self.transaction

    def _app_run(self, *args, **kwargs):
        if not self.transaction.message or self.transaction._committed:
            self.consume(*args, **kwargs)
            self.app_function(self.transaction, *self.app_function_arglist)
        self.commit()

    def _app_run_loop(self, *args, on_no_message=None, **kwargs):
        while not self._shutdown:
            try:
                LOGGER.debug('Running app function...')
                self._app_run(*args, **kwargs)
            except NoMessageError:
                self.producer.poll(0)
                LOGGER.info('No messages!')
                if on_no_message:
                    on_no_message()
            except GracefulTransactionFailure:
                LOGGER.info("Graceful transaction failure; retrying message with a new transaction...")
                self._abort_active_transaction()
            except FatalTransactionFailure:
                LOGGER.info("Fatal transaction failure; recreating the producer and retrying message...")
                self._abort_active_transaction()

    def kafka_cleanup(self):
        """ Public method in the rare cases where you need to do some cleanup on the consumer object manually. """
        shutdown_cleanup(consumer=self.consumer)

    def _app_shutdown(self):
        LOGGER.info('App is shutting down...')
        self._shutdown = True
        self.transaction.abort_active_transaction()
        self.kafka_cleanup()

    def run(self, *args, health_path='/tmp', **kwargs):
        """
        # as_loop is really only for rare apps that don't follow the typical consume-looping behavior
        (ex: async apps) and don't seem to raise out of the True loop as expected.
        """
        LOGGER.info('RUN initialized!')
        with open(f'{health_path}/health', 'w') as health_file:
            health_file.write('Healthy')
        try:
            self._app_run_loop(*args, **kwargs)
        except SignalRaise:
            LOGGER.info('Shutdown requested!')
        except Exception as e:
            LOGGER.error(e)
            if self.metrics_manager:
                log_and_raise_error(self.metrics_manager, e)
        finally:
            self._app_shutdown()
            try:
                remove(f'{health_path}/health')
            except:
                pass



from .gtfo_table_app import GtfoTableApp, TableTransaction
from .gtfo_app import NoMessageError, KafkaException, handle_kafka_exception
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from confluent_kafka import TopicPartition
import logging
from datetime import datetime


LOGGER = logging.getLogger(__name__)


class AutoBatchTableTransaction(TableTransaction):
    def __init__(self, producer, consumer, app_changelog_topic, app_tables, metrics_manager=None, message=None, auto_consume=True, timeout=5, parent_app=None):
        self._commit_offsets = {}
        self.message_count = 0
        super().__init__(producer, consumer, app_changelog_topic=app_changelog_topic, app_tables=app_tables, metrics_manager=metrics_manager, message=message, auto_consume=auto_consume, timeout=timeout, parent_app=parent_app)

    def _refresh_transaction_fields(self):
        super()._refresh_transaction_fields()
        self._commit_offsets = {}
        self.message_count = 0

    def consume(self):
        super().consume()
        self._commit_offsets[self.partition()] = {"topic": self.topic(), "partition": self.partition(), "offset": self.offset()}
        self.message_count += 1

    def commit(self, mark_committed=True):
        # TODO: Confirm there aren't commits from old partition assignments (not sure how that's handled...it may not rebalance
        #  while doing a transaction?)
        try:
            LOGGER.info('Committing Transaction...')
            offsets_to_commit = [TopicPartition(msg['topic'], msg['partition'], msg['offset'] + 1) for msg in
                                 self._commit_offsets.values()]
            if offsets_to_commit:
                LOGGER.debug(f'Offsets to commit: {offsets_to_commit}')
                self._init_transaction()
            if self._active_transaction:
                LOGGER.info(f'Committing {self.message_count} messages...')
                self.producer.send_offsets_to_transaction(offsets_to_commit, self.consumer.consumer_group_metadata())
                self.producer.commit_transaction(30)
                self.producer.poll(0)  # TODO: experiment with flushing? I think it does it by default
                LOGGER.info('All partition transactions committed!')
                self._committed = mark_committed
                self._commit_offsets = {}
                self.message_count = 0
                self._retry_offsets = {}
                if self._committed:
                    self._active_transaction = False
            self._table_write()
        except KafkaException as kafka_error:
            handle_kafka_exception(kafka_error)


class GtfoAutoBatchTableApp(GtfoTableApp):
    def __init__(self, app_function, consume_topic, produce_topic_schema_dict=None, transaction_type=AutoBatchTableTransaction,
                 app_function_arglist=None, metrics_manager=None, schema_registry=None, cluster_name=None, consumer=None, producer=None,
                 time_elapse_max_seconds=None, consume_max_count=None):

        self._time_elapse_start = None

        if time_elapse_max_seconds is None:
            time_elapse_max_seconds = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_TIME_SECONDS'])
        if consume_max_count is None:
            consume_max_count = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_COUNT'])

        self._time_elapse_max_seconds = time_elapse_max_seconds
        self._consume_max_count = consume_max_count

        super().__init__(
            app_function, consume_topic, produce_topic_schema_dict, transaction_type=transaction_type,
            app_function_arglist=app_function_arglist, metrics_manager=metrics_manager, schema_registry=schema_registry, cluster_name=cluster_name, consumer=consumer, producer=producer)

    def prepare_transaction(self, *args, **kwargs):
        return super().consume(*args, auto_consume=False, **kwargs)

    def _max_consume_time_continue(self):
        if self._time_elapse_max_seconds:
            seconds_elapsed = datetime.now().timestamp() - self._time_elapse_start
            return seconds_elapsed < self._time_elapse_max_seconds
        return True

    def _max_consume_count_continue(self, consume_multiplier=1):
        if self._consume_max_count:
            return self.transaction.message_count < (self._consume_max_count * consume_multiplier)
        return True

    def _keep_consuming(self, consume_multiplier=None):
        if not consume_multiplier:
            consume_multiplier = 1
        return not self._shutdown and self._max_consume_count_continue(consume_multiplier=consume_multiplier) and self._max_consume_time_continue()

    def _check_pending_recovery_status(self):
        recovery_elapsed_time = int(datetime.timestamp(datetime.now())) - self._recovery_time_start
        LOGGER.debug(f'Current pending recovery partition statuses before refreshing: {self._pending_table_recoveries}')
        recovery_status = {p: {'table_offset': self.transaction._table_offset(partition=p), 'highwater': self._pending_table_recoveries[p]['watermarks'][1]} for p in self._pending_table_recoveries.keys()}
        current_recovery_offsets_remaining = {p: recovery_status[p]['highwater'] - recovery_status[p]['table_offset'] for p in recovery_status}
        total_remaining_offsets = sum(current_recovery_offsets_remaining.values())
        offsets_processed = self._recovery_offsets_remaining - total_remaining_offsets
        self._recovery_offsets_handled += offsets_processed
        self._recovery_offsets_remaining = total_remaining_offsets
        recovery_rate_offsets_sec = self._recovery_offsets_handled / recovery_elapsed_time
        remaining_time_mins = round((self._recovery_offsets_remaining / recovery_rate_offsets_sec) / 60, 2)
        LOGGER.debug(f'RECOVERY - offsets processed this iteration: {offsets_processed}')
        LOGGER.debug(f'RECOVERY - current process rate: {int(recovery_rate_offsets_sec)} offsets/sec')
        LOGGER.info(f'RECOVERY - {self._recovery_offsets_handled} TOTAL OFFSETS PROCESSED OVER {round(recovery_elapsed_time / 60, 2)} minutes')
        LOGGER.info(f'RECOVERY - ESTIMATED TIME REMAINING: {remaining_time_mins} minutes')
        LOGGER.info(f'RECOVERY - OFFSETS REMAINING: {current_recovery_offsets_remaining}')
        for p in recovery_status:
            if recovery_status[p]['highwater'] - (recovery_status[p]['table_offset'] + 2) <= 0:  # +2 comes from: +1 due to table tracking offset it has up to, and another +1 due to last offset expecting to be a transaction marker
                LOGGER.info(f'Offset delta for p{p} is within an expected tolerable range of the changelog highwater; it is safe to assume transactional markers are the remaining delta.')
                self._set_table_offset_to_latest(p)

    def _table_recovery_loop(self, checks=2, consume_multiplier=10):
        while self._pending_table_recoveries and checks:
            LOGGER.info(f'Consuming from changelog partitions: {list(self._active_table_changelog_partitions.keys())}')
            self.prepare_transaction()
            if self._time_elapse_max_seconds:
                self._time_elapse_start = datetime.now().timestamp()
            LOGGER.info(f'Processing up to {self._consume_max_count * consume_multiplier} messages for up to {self._time_elapse_max_seconds} seconds!')
            # NOTE: no transaction commits since its just consuming from changelog and writing to the table, we dont care about the consumer group offset
            try:
                while self._keep_consuming(consume_multiplier=consume_multiplier) and self._pending_table_recoveries and checks:
                    self.transaction.consume()
                    self.transaction._update_table_entry_from_changelog()
            except NoMessageError:  # NOTE: this does not necc mean no messages were pulled at all in the batch, just that there's now none left!
                checks -= 1
                LOGGER.debug(f'No further changelog messages, checks remaining: {checks}')
            LOGGER.info(f'Consumed {self.transaction.message_count} changelog messages')
            self.transaction._table_write(recovery_multiplier=consume_multiplier)  # relay transaction's cached writes to the table's write cache
            self._check_pending_recovery_status()
        self.commit_tables()
        self._check_pending_recovery_status()
        if not checks:
            for p in list(self._pending_table_recoveries.keys()):  # list stops dict iter size change exception
                self._set_table_offset_to_latest(p)

    def _app_run(self, *args, **kwargs):
        LOGGER.info(f'Consuming from partitions: {list(self._active_primary_partitions.keys())}')
        self.prepare_transaction(*args, **kwargs)
        if self._time_elapse_max_seconds:
            self._time_elapse_start = datetime.now().timestamp()
        LOGGER.info(
            f'Processing up to {self._consume_max_count} messages for up to {self._time_elapse_max_seconds} seconds!')
        try:
            while self._keep_consuming():
                self.transaction.consume()
                self.app_function(self.transaction, *self.app_function_arglist)
        except NoMessageError:
            self.producer.poll(0)
            LOGGER.info('No messages!')
        if any(self.transaction._pending_table_offset_increase.values()) or self.transaction._commit_offsets:
            self.transaction.commit()
            self.check_table_commits()
        else:
            self.commit_tables()



from nubium_utils.confluent_utils.producer_utils import get_producers
from nubium_utils.confluent_utils.confluent_configs import init_schema_registry_configs, env_vars
from nubium_utils.yaml_parser import load_yaml_fp


def gtfo_producer(topic_schema_dict):
    """
    This assumes all produce topics are on the same cluster, too lazy to add a check for now
    """
    cluster_name = load_yaml_fp(env_vars()['NU_TOPIC_CONFIGS_YAML'])[list(topic_schema_dict.keys())[0]]['cluster']
    return get_producers(topic_schema_dict, cluster_name, init_schema_registry_configs())



from unittest import TestCase
from unittest.mock import patch
from nubium_utils.dict_manipulator import DictManipulator
from copy import deepcopy


class TestDictManipulator(TestCase):
    def tearDown(self):
        self.env_patch.stop()

    def setUp(self):
        self.env = {}
        self.env_patch = patch.dict('os.environ', self.env)
        self.env_patch.start()

        self.junk_func = lambda x: x + '_' if x else ''

        self.fields = {
            "truths": {
                "email_address": '',
                "first_name": 'yay',
                "last_name": 'nay',
                "address": {
                    "address_street": '',
                    "address_city": 'derp??',
                    "address_state": '',
                    "address_postal_code": ''
                },
                "country_name": '',
                "country_code": '',
                "company": '',
                "department": '',
                "industry": '',
                "job_title": 'neat'
            },
            "not_truths": {
                "blah": 'woo-hoo.com',
                "blah2": 'he-he.com'
            }
        }

        self.fields_junk_edit = {
            "truths": {
                "email_address": '',
                "first_name": 'yay_',
                "last_name": 'nay',
                "address": {
                    "address_street": '',
                    "address_city": 'derp??_',
                    "address_state": '',
                    "address_postal_code": ''
                },
                "country_name": '',
                "country_code": '',
                "company": '',
                "department": '',
                "industry": '',
                "job_title": 'neat_'
            },
            "not_truths": {
                "blah": 'woo-hoo.com_',
                "blah2": 'he-he.com'
            }
        }

        self.fields_subset_mask = {
            "truths": {
                "email_address": None,
                "first_name": None,
                "job_title": None,
                "address": {
                    "address_city": None,
                    "address_state": None
                }
            },
            "not_truths": {
                'blah': None
            }
        }

        self.fields_subset_populated = {
            "truths": {
                "email_address": '',
                "first_name": 'yay',
                "job_title": 'neat',
                "address": {
                    "address_city": 'derp??',
                    "address_state": ''
                }
            },
            "not_truths": {
                'blah': 'woo-hoo.com'
            }
        }

        self.dict_manipulator = DictManipulator(self.fields)

    def test__apply_funcs(self):
        self.assertEqual(self.dict_manipulator._apply_funcs([lambda x: x + 'pp', lambda y: y + 'le'], 'a'), 'apple')
        self.assertEqual(self.dict_manipulator._apply_funcs(self.junk_func, 'a'), 'a_')

    def test_apply_func_dict_to_field(self):
        def dumb_func_1(word, junk1=''):
            return word + junk1
        def dumb_func_2(word, junk2=''):
            return word + junk2
        self.dict_manipulator.apply_func_dict_to_field('truths.first_name',
                                                       {dumb_func_1: {'junk1': '_'}, dumb_func_2: {'junk2': 'z'}})
        self.assertEqual(self.dict_manipulator['truths']['first_name'], 'yay_z')

    def test__apply_funcs_to_mask_selection(self):
        self.dict_manipulator._apply_funcs_to_mask_selection(self.fields, self.fields_subset_mask, self.junk_func)
        self.assertDictEqual(self.fields, self.fields_junk_edit)

    def test_apply_funcs_to_mask_selection(self):
        self.dict_manipulator.apply_funcs_to_mask_selection(self.fields_subset_mask, self.junk_func)
        self.assertDictEqual(self.dict_manipulator, self.fields_junk_edit)

    def test__check_if_blanked(self):
        self.assertEqual(self.dict_manipulator._check_if_nulled('woo', ''), 'woo')
        self.assertEqual(self.dict_manipulator._check_if_nulled('', 'wee'), 'wee')
        self.assertEqual(self.dict_manipulator._check_if_nulled('', ''), '')
        self.assertEqual(self.dict_manipulator._check_if_nulled('woo', 'wee'), 'wee')

    def test__compare_and_overwrite_blanked_values(self):
        self.full_fields_edited = deepcopy(self.fields)
        self.full_fields_edited['not_truths']['blah'] = ''

        self.dict_manipulator._compare_and_overwrite_nulled_values(self.full_fields_edited, self.fields)
        self.assertDictEqual(self.full_fields_edited, self.fields)

    def test_compare_and_overwrite_blanked_values(self):
        self.dict_manipulator['not_truths']['blah'] = ''
        self.dict_manipulator.compare_and_overwrite_nulled_values(self.fields)
        self.assertDictEqual(self.dict_manipulator, self.fields)

    def test__get_populated_mask(self):
        self.assertDictEqual(
            self.dict_manipulator._view_populated_mask(self.fields, self.fields_subset_mask),
            self.fields_subset_populated)

    def test_get_populated_mask(self):
        self.assertDictEqual(self.dict_manipulator.view_populated_mask(self.fields_subset_mask), self.fields_subset_populated)

    def test_get_dict_with_dropped_keys(self):
        output = {
            "truths": {
                "first_name": 'yay',
                "last_name": 'nay',
                "address": {
                    "address_street": '',
                    "address_city": 'derp??',
                    "address_state": '',
                    "address_postal_code": ''
                },
                "country_name": '',
                "country_code": '',
                "company": '',
                "department": '',
                "industry": '',
                "job_title": 'neat'
            },
            "not_truths": {
                "blah2": 'he-he.com'
            }
        }
        self.assertDictEqual(self.dict_manipulator.view_while_ignoring_fields([('truths', 'email_address'),
                                                                               ('not_truths', 'blah')]),
                         output)



from confluent_kafka import TopicPartition, KafkaException
from nubium_utils.custom_exceptions import NoMessageError
from nubium_utils.confluent_utils.confluent_configs import init_schema_registry_configs, env_vars
from .sqlite_utils import SqliteGtfo
from .gtfo_app import GtfoApp, Transaction
import logging
from time import sleep
from json import dumps, loads
from datetime import datetime
from collections import deque
from copy import deepcopy
# TODO: consider using orjson for changelog speed increases?
# from orjson import dumps, loads
# changelog_schema = {"type": "bytes"}
changelog_schema = {"type": "string"}


LOGGER = logging.getLogger(__name__)


class PartitionsAssigned(Exception):
    def __init__(self):
        pass


class TableTransaction(Transaction):
    def __init__(self, producer, consumer, app_changelog_topic, app_tables, metrics_manager=None, message=None, auto_consume=True, timeout=5, parent_app=None):
        self.app_changelog_topic = app_changelog_topic
        self.app_tables = app_tables
        self._is_not_changelog_message = True
        self._pending_table_writes = {p: {} for p in self.app_tables}
        self._pending_table_offset_increase = {p: 0 for p in self.app_tables}
        super().__init__(producer, consumer, metrics_manager=metrics_manager, message=message, auto_consume=auto_consume, timeout=timeout, parent_app=parent_app)

    def _refresh_transaction_fields(self):
        super()._refresh_transaction_fields()
        self._is_not_changelog_message = True
        self._pending_table_writes = {p: {} for p in self.app_tables}
        self._pending_table_offset_increase = {p: 0 for p in self.app_tables}

    def read_table_entry(self):
        pending_update = self._pending_table_writes.get(self.partition(), {}).get(self.key())
        if pending_update:
            return deepcopy(pending_update)
        return self.app_tables[self.partition()].read(self.key())

    def _update_pending_table_writes(self, value):
        self._pending_table_writes[self.partition()][self.key()] = value
        if self._is_not_changelog_message:
            self._pending_table_offset_increase[self.partition()] += 1
        else:
            self._pending_table_offset_increase[self.partition()] += self.offset() - self._table_offset()

    def update_table_entry(self, value):
        self._update_pending_table_writes(value)
        self._update_changelog()

    def delete_table_entry(self):
        self._update_pending_table_writes('-DELETED-')
        self._update_changelog()

    def _update_changelog(self):
        if self._is_not_changelog_message and (pending_write := self._pending_table_writes.get(self.partition(), {}).get(self.key())):
            LOGGER.debug('Updating changelog topic...')
            self.produce(dict(
                topic=self.app_changelog_topic,
                key=self.key(),
                value=dumps(pending_write),
                partition=self.partition()
            ))
            self.producer.poll(0)

    def _update_table_entry_from_changelog(self):
        self._is_not_changelog_message = False  # so we dont produce a message back to the changelog
        self.update_table_entry(loads(self.value()))

    def _table_write(self, recovery_multiplier=None):
        for p, msgs in self._pending_table_writes.items():
            if msgs:
                table = self.app_tables[p]
                LOGGER.debug(f'Finalizing table entry batch write of {len(msgs)} records for table {table.table_name}')
                LOGGER.debug(f'Table {table.table_name} offset before write: {table.offset}, expected after write: {self._table_offset(partition=p) + int(self._is_not_changelog_message)}')
                table.set_offset(self._table_offset(partition=p) + int(self._is_not_changelog_message))  # the +1 is because the transaction causes an extra offset at the end (per trans + partition)
                table.write_batch(msgs)
                table.commit_and_cleanup_if_ready(recovery_multiplier=recovery_multiplier)
                self._pending_table_writes[p] = {}
                self._pending_table_offset_increase[p] = 0

    def _table_offset(self, partition=None):
        """
        Partition arg generally only used by subclasses
        """
        if partition is None:
            partition = self.partition()
        value = self.app_tables[partition].offset
        value = int(value) if value else 0
        return value + self._pending_table_offset_increase.get(partition, 0)

    def commit(self, mark_committed=True):
        super().commit(mark_committed=mark_committed)
        self._table_write()
        LOGGER.info('Transaction Committed!')


class GtfoTableApp(GtfoApp):
    def __init__(self, app_function, consume_topic, produce_topic_schema_dict=None, transaction_type=TableTransaction,
                 app_function_arglist=None, metrics_manager=None, schema_registry=None, cluster_name=None, consumer=None, producer=None):
        self.changelog_topic = f"{env_vars()['NU_APP_NAME']}__changelog"
        self.tables = {}
        self._active_primary_partitions = {}
        self._paused_primary_partitions = {}
        self._active_table_changelog_partitions = {}
        self._pending_table_db_assignments = deque()
        self._pending_table_recoveries = {}

        if not produce_topic_schema_dict:
            produce_topic_schema_dict = {}
        if self.changelog_topic not in produce_topic_schema_dict:
            produce_topic_schema_dict.update({self.changelog_topic: changelog_schema})
        if not cluster_name:
            cluster_name = self._get_cluster_name(consume_topic)
        if not schema_registry:
            schema_registry = init_schema_registry_configs()
        if not consumer:
            consumer = self._set_table_consumer(consume_topic, schema_registry, cluster_name=cluster_name)

        self.consume_topic = consume_topic

        # Recovery est. time remaining metadata
        self._recovery_time_start = 0
        self._recovery_offsets_remaining = 0
        self._recovery_offsets_handled = 0

        super().__init__(
            app_function, self.consume_topic, produce_topic_schema_dict, transaction_type=transaction_type,
            app_function_arglist=app_function_arglist, metrics_manager=metrics_manager, schema_registry=schema_registry, cluster_name=cluster_name, consumer=consumer, producer=producer)

    def _set_table_consumer(self, topic, schema_registry, default_schema=None, cluster_name=None):
        if isinstance(topic, str):
            topic = [topic]
        consumer = self._get_transactional_consumer(topic, schema_registry, cluster_name, default_schema, False)
        consumer.subscribe(topic, on_assign=self._partition_assignment, on_revoke=self._partition_unassignment, on_lost=self._partition_unassignment)
        LOGGER.debug('Consumer initialized.')
        return consumer

    def _table_close(self, partitions=None):
        interrupt = None
        full_shutdown = False
        if not partitions or self._shutdown:
            partitions = list(self.tables.keys())
            full_shutdown = True
        LOGGER.debug(f'Table - closing connections for partitions {partitions}')
        for p in partitions:
            try:
                self.tables[p].close()
                sleep(.1)
                LOGGER.debug(f'p{p} table connection closed.')
                del self.tables[p]
            except KeyError:
                if not full_shutdown:
                    LOGGER.debug(
                        f'Table p{p} did not seem to be mounted and thus could not unmount,'
                        f' likely caused by multiple rebalances in quick succession.'
                        f' This is unliklely to cause issues as the client is in the middle of adjusting itself, '
                        f' but should be noted.')
            except Exception as e:
                LOGGER.debug(f'Interrupt received during table db closing, {e}')
                interrupt = e
        LOGGER.info(f'Table - closed connections for partitions {partitions}')
        if interrupt:  # ensure all table cleanup happens
            LOGGER.info('Continuing with exception interrupt raised during table closing')
            raise interrupt

    def _partition_assignment(self, consumer, add_partition_objs):
        """
        Called every time a rebalance happens and handles table assignment and recovery flow.
        NOTE: rebalances pass relevant partitions per rebalance call which can happen multiple times, especially when
        multiple apps join at once; we have objects to track all updated partitions received during the entire rebalance.
        NOTE: confluent-kafka expects this method to have exactly these two arguments ONLY
        NOTE: _partition_assignment will ALWAYS be called (even when no new assignments are required) after _partition_unassignment.
            """
        if not self._shutdown:
            LOGGER.debug('Rebalance Triggered - Assigment')
            self.transaction.abort_active_transaction()
            partitions = {p_obj.partition: p_obj for p_obj in add_partition_objs}
            if add_partition_objs:
                LOGGER.info(f'Rebalance - Assigning additional partitions: {list(partitions.keys())}')
                self._paused_primary_partitions.update(partitions)
                self._pending_table_db_assignments.extend(list(partitions.keys()))
                self._pending_table_recoveries.update({p: None for p in partitions.keys()})
                self.consumer.incremental_assign(add_partition_objs)
                self.consumer.pause(add_partition_objs)
                raise PartitionsAssigned  # want to interrupt what it was doing
            else:
                LOGGER.debug(f'No new partitions assigned, skipping rebalance handling')

    def _partition_unassignment(self, consumer, drop_partition_objs):
        """
        NOTE: confluent-kafka expects this method to have exactly these two arguments ONLY
        NOTE: _partition_assignment will always be called (even when no new assignments are required) after _partition_unassignment.
        """
        LOGGER.info('Rebalance Triggered - Unassigment')
        self.transaction.abort_active_transaction()
        partitions = [p_obj.partition for p_obj in drop_partition_objs]
        self._table_close([p for p in self.tables.keys() if p in partitions])
        unassign_active_changelog = {p: p_obj for p, p_obj in self._active_table_changelog_partitions.items() if p in partitions}
        full_unassign = {self.changelog_topic: list(unassign_active_changelog.keys()), self.consume_topic: partitions}
        if self._pending_table_recoveries:
            full_unassign.update({self.changelog_topic: list(unassign_active_changelog.keys())}),

        LOGGER.info(f'Unassigning partitions:\n{full_unassign}')
        self.consumer.incremental_unassign(list(unassign_active_changelog.values()) + drop_partition_objs)

        for var in ['_active_primary_partitions', '_paused_primary_partitions', '_active_table_changelog_partitions', '_pending_table_recoveries']:
            self.__setattr__(var, {k: v for k, v in self.__getattribute__(var).items() if k not in partitions})
            LOGGER.debug(f'{var} after unassignment: {list(self.__getattribute__(var).keys())}')
        self._pending_table_db_assignments = deque([i for i in self._pending_table_recoveries if i not in partitions])
        LOGGER.debug(f'_pending_table_db_assignments after unassignment: {self._pending_table_db_assignments}')
        LOGGER.info('Unassignment Complete!')

    def _pause_active_primary_partitions(self):
        self.consumer.pause(list(self._active_primary_partitions.values()))
        self._paused_primary_partitions.update(self._active_primary_partitions)
        self._active_primary_partitions = {}

    def _resume_active_primary_partitions(self):
        self.consumer.resume(list(self._paused_primary_partitions.values()))
        self._active_primary_partitions.update(self._paused_primary_partitions)
        self._paused_primary_partitions = {}

    def _cleanup_and_resume_app_loop(self):
        if self._active_table_changelog_partitions:
            LOGGER.info(f'unassigning changelog partitions: {list(self._active_table_changelog_partitions.values())}')
            self.consumer.incremental_unassign(list(self._active_table_changelog_partitions.values()))
            self._active_table_changelog_partitions = {}
        LOGGER.debug(f'Resuming consumption for paused topic partitions:\n{list(self._paused_primary_partitions.keys())}')
        self._resume_active_primary_partitions()
        LOGGER.info(f'Continuing normal consumption loop for partitions {list(self._active_primary_partitions.keys())}')

    def _get_changelog_watermarks(self, p_pobj_dict):
        """
        Note: this is a separate function since it requires the consumer to communicate with the broker
        """
        return {p: {'watermarks': self.consumer.get_watermark_offsets(p_obj, timeout=8), 'partition_obj': p_obj} for p, p_obj in p_pobj_dict.items()}

    def _table_recovery_set_offset_seek(self):
        """ Refresh the offsets of recovery partitions to account for updated recovery states during rebalancing """
        for p, offsets in self._pending_table_recoveries.items():
            new_offset = self._pending_table_recoveries[p]['table_offset_recovery']
            low_mark = self._pending_table_recoveries[p]['watermarks'][0]
            if low_mark > new_offset:  # handles offsets that have been removed/compacted. Should never happen, but ya know
                LOGGER.info(
                    f'p{p} table has an offset ({new_offset}) less than the changelog lowwater ({low_mark}), likely due to retention settings. Setting {low_mark} as offset start point.')
                new_offset = low_mark
                self.tables[p].set_offset(new_offset - 1)  # -1 since the table marks the latest offset it has, not which offset is next
            high_mark = self._pending_table_recoveries[p]['watermarks'][1]
            LOGGER.debug(f'p{p} table has an offset delta of {high_mark - new_offset}')
            self._pending_table_recoveries[p]['partition_obj'].offset = new_offset
        self._recovery_offsets_remaining = sum([self._pending_table_recoveries[p]['watermarks'][1] - self._pending_table_recoveries[p]['partition_obj'].offset for p in self._pending_table_recoveries])

    def _refresh_pending_table_recoveries(self):
        """
        confirms new recoveries and removes old ones if not applicable anymore
        """
        if env_vars()['NU_SKIP_TABLE_RECOVERY'] == 'true':  # DEBUGGING PURPOSES ONLY
            self._pending_table_recoveries = {}
        else:
            partition_watermarks = self._get_changelog_watermarks({p: TopicPartition(topic=self.changelog_topic, partition=int(p)) for p, d in self._pending_table_recoveries.items()})
            for partition, data in partition_watermarks.items():
                watermarks = data['watermarks']
                if watermarks[0] != watermarks[1]:
                    table_offset = self.tables[partition].offset
                    LOGGER.info(f'(lowwater, highwater) [table_offset] for changelog p{partition}: {watermarks}, [{table_offset}]')
                    if table_offset < watermarks[1]:
                        data['table_offset_recovery'] = table_offset
            self._pending_table_recoveries = {k: v for k, v in partition_watermarks.items() if v.get('table_offset_recovery') is not None}
            LOGGER.debug(f'Remaining recoveries after offset refresh check: {self._pending_table_recoveries}')
            self._table_recovery_set_offset_seek()

    def _table_db_init(self, partition):
        self.tables[partition] = SqliteGtfo(f'p{partition}')

    def _table_db_assignments(self):
        while self._pending_table_db_assignments:
            partition = self._pending_table_db_assignments.popleft()
            if partition not in self.tables:
                sleep(.1)  # try to slow down pvc stuff a little
                self._table_db_init(partition)

    def _set_table_offset_to_latest(self, partition):
        LOGGER.debug(f'Setting table offset p{partition} to latest')
        self.tables[partition].offset = self._pending_table_recoveries[partition]['watermarks'][1]
        self.tables[partition].commit()
        del self._pending_table_recoveries[partition]
        LOGGER.info(f'table p{partition} fully recovered!')

    def _table_recovery_loop(self, checks=3):
        while checks and self._pending_table_recoveries:
            try:
                LOGGER.info(f'Consuming from changelog partitions: {list(self._active_table_changelog_partitions.keys())}')
                self.consume()
                self.transaction._update_table_entry_from_changelog()
                # NOTE: no commit here since its just writing to the table
                p = self.transaction.partition()
                LOGGER.info(f"transaction_offset - {self.transaction.offset() + 2}, watermark - {self._pending_table_recoveries[p]['watermarks'][1]}")
                self.transaction._table_write()
                if self._pending_table_recoveries[p]['watermarks'][1] - (self.transaction.offset() + 2) <= 0:
                    self._set_table_offset_to_latest(p)
                self.tables[p].commit_and_cleanup_if_ready()
            except NoMessageError:
                checks -= 1
                LOGGER.debug(f'No changelog messages, checks remaining: {checks}')
        self.commit_tables()
        if not checks:
            for p in list(self._pending_table_recoveries.keys()):  # list stops dict iter size change exception
                self._set_table_offset_to_latest(p)

    def _throwaway_poll(self):
        """ Have to poll (changelog topic) after first assignment to it to allow seeking """
        LOGGER.debug("Performing throwaway poll to allow assignments to properly initialize...")
        try:
            self.consume(timeout=8)
        except NoMessageError:
            pass

    def _assign_recovery_partitions(self):
        # TODO: maybe add additional check/inclusion of assignment based on c.assignment to avoid desync? (hasnt happened yet though)
        LOGGER.debug(f'Preparing changelog table recovery partition assignments...')
        to_assign = {p: self._pending_table_recoveries[p]['partition_obj'] for p in
                     self._pending_table_recoveries.keys() if p not in self._active_table_changelog_partitions}
        LOGGER.info(f'Assigning changelog partitions {to_assign}')
        self._active_table_changelog_partitions.update(to_assign)
        self.consumer.incremental_assign(list(to_assign.values()))  # strong assign due to needing to seek
        LOGGER.debug(f'pending table recoveries before recovery attempt: {list(self._pending_table_recoveries.keys())}')
        LOGGER.debug(f'assigned recoveries before recovery attempt (should match pending now): {self._active_table_changelog_partitions}')

    def _table_recovery_start(self):
        try:
            for partition_info in self._pending_table_recoveries.values():
                self.consumer.seek(partition_info['partition_obj'])
            self._table_recovery_loop()
        except KafkaException as kafka_error:
            if 'Failed to seek to offset' in kafka_error.args[0].str():
                LOGGER.debug('Running a consumer poll to allow seeking to work on the changelog partitions...')
                self._throwaway_poll()
                self._table_recovery_start()
            else:
                raise

    def _table_and_recovery_manager(self):
        self._recovery_time_start = int(datetime.timestamp(datetime.now()))
        try:
            self._table_db_assignments()
            self._refresh_pending_table_recoveries()
            if self._pending_table_recoveries:
                self._pause_active_primary_partitions()
                while self._pending_table_recoveries:
                    self._assign_recovery_partitions()
                    LOGGER.info('BEGINNING TABLE RECOVERY PROCEDURE')
                    self._table_recovery_start()
                    if self._pending_table_recoveries:
                        self._refresh_pending_table_recoveries()
                LOGGER.info("TABLE RECOVERY COMPLETE!")
            else:
                LOGGER.info('No table recovery required!')
            self._cleanup_and_resume_app_loop()
        except PartitionsAssigned:
            LOGGER.info('Rebalance triggered while recovering tables. Will resume recovery after, if needed.')
            self._table_and_recovery_manager()

    def consume(self, *args, **kwargs):
        return super().consume(*args, app_changelog_topic=self.changelog_topic, app_tables=self.tables, **kwargs)

    def check_table_commits(self, recovery_multiplier=None):
        if not recovery_multiplier:
            recovery_multiplier = 1
        for table in self.tables.values():
            table.commit_and_cleanup_if_ready(recovery_multiplier=recovery_multiplier)

    def commit_tables(self):
        for table in self.tables.values():
            table.commit()

    def _app_run_loop(self, *args, **kwargs):
        LOGGER.info(f'Consuming from partitions: {list(self._active_primary_partitions.keys())}')
        while not self._shutdown:
            try:
                super()._app_run_loop(*args, on_no_message=self.commit_tables, **kwargs)
                self.check_table_commits()
            except PartitionsAssigned:
                self._table_and_recovery_manager()

    def _app_shutdown(self):
        super()._app_shutdown()
        self._table_close()



from unittest import TestCase

from nubium_utils.hashing_utils import hash_record


class TestRecordHasher(TestCase):

    def setUp(self) -> None:
        self.test_record = {
            "desired_field_1": "value_1",
            "desired_field_2": "value_2",
            "undesired_field_3": "value_3"
        }
        self.desired_field_mask = {
            "desired_field_1": True,
            "desired_field_2": True,
            "undesired_field_3": False
        }
        self.expected_hex_digest = '50cb97c5e1a0078bad72104eb389b65b'

    def test_dictionary_hash(self):
        """Function should return a hash of a dictionary using the desired fields"""
        record_hash = hash_record(self.test_record, self.desired_field_mask)
        assert self.expected_hex_digest == record_hash.hexdigest()

    def test_ignores_unincluded_fields(self):
        """Shouldn't include fields that we don't want in the hash"""
        record_hash = hash_record(self.test_record, self.desired_field_mask)
        hex_digest_using_all_fields = '8c7855a8a95074460d09ea7a00ecff7f'
        assert hex_digest_using_all_fields != record_hash.hexdigest()

    def test_repeatable_hash(self):
        """Hash function should return the same output given the same input"""
        record_hash_1 = hash_record(self.test_record, self.desired_field_mask)
        record_hash_2 = hash_record(self.test_record, self.desired_field_mask)

        assert record_hash_1.hexdigest() == record_hash_2.hexdigest()

    def test_nested_fields(self):
        test_record = {
            "desired_field_1": "value_1",
            "desired_field_2": {
                "desired_field_3": "value_3",
                "undesired_field_4": "value_4",
                "desired_field_5": {
                    "desired_field_6": "value_6"
                }
            },
            "undesired_field_7": "value_7"
        }
        test_map = {
            "desired_field_1": True,
            "desired_field_2": {
                "desired_field_3": True,
                "undesired_field_4": False,
                "desired_field_5": {
                    "desired_field_6": True
                }
            },
            "undesired_field_7": False
        }
        output_record = hash_record(test_record, test_map)
        expected_output = '63f8a9b3e2836f6f469154d59a7270d2'
        assert output_record.hexdigest() == expected_output

    def test_list_field(self):
        test_record = {
            "field_1": ["value_1", "value_2", "value_3"]
        }
        test_map = {"field_1": True}
        output_hash = hash_record(test_record, test_map)
        expected_output = '81cc8e59f248a7ec3a94bd105c48619c'
        assert output_hash.hexdigest() == expected_output




from confluent_kafka import TopicPartition
from nubium_utils.custom_exceptions import NoMessageError
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
import logging
from datetime import datetime
from .gtfo_app import GtfoApp, Transaction


LOGGER = logging.getLogger(__name__)


class AutoBatchTransaction(Transaction):
    def __init__(self, producer, consumer, metrics_manager=None, auto_consume=False, timeout=None, parent_app=None):
        self._commit_offsets = {}
        self.message_count = 0
        super().__init__(producer, consumer, metrics_manager=metrics_manager, auto_consume=auto_consume, timeout=timeout,
                         message=None, parent_app=parent_app)

    def _refresh_transaction_fields(self):
        super()._refresh_transaction_fields()
        self._commit_offsets = {}
        self.message_count = 0

    def consume(self):
        super().consume()
        self._commit_offsets[self.partition()] = {"topic": self.topic(), "partition": self.partition(), "offset": self.offset()}
        self.message_count += 1

    def commit(self, mark_committed=True):
        # TODO: Confirm there aren't commits from old partition assignments (not sure how that's handled...it may not rebalance
        #  while doing a transaction?)
        offsets_to_commit = [TopicPartition(msg['topic'], msg['partition'], msg['offset'] + 1) for msg in
                             self._commit_offsets.values()]
        if offsets_to_commit:
            LOGGER.debug(f'Offsets to commit: {offsets_to_commit}')
            self._init_transaction()
        if self._active_transaction:
            LOGGER.info(f'Committing {self.message_count} messages...')
            self.producer.send_offsets_to_transaction(offsets_to_commit, self.consumer.consumer_group_metadata())
            self.producer.commit_transaction(30)
            self.producer.poll(0)
            LOGGER.debug('All partition transactions committed!')
            self._committed = mark_committed
            self._commit_offsets = {}
            self.message_count = 0
            self._retry_offsets = {}
            if self._committed:
                self._active_transaction = False


class GtfoAutoBatchApp(GtfoApp):
    def __init__(self, app_function, consume_topics_list, produce_topic_schema_dict=None,
                 transaction_type=AutoBatchTransaction,
                 app_function_arglist=None, metrics_manager=None, schema_registry=None, cluster_name=None,
                 consumer=None, producer=None, time_elapse_max_seconds=None, consume_max_count=None):
        self._time_elapse_start = None

        if not time_elapse_max_seconds:
            time_elapse_max_seconds = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_TIME_SECONDS'])
        if not consume_max_count:
            consume_max_count = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_COUNT'])

        self._time_elapse_max_seconds = time_elapse_max_seconds
        self._consume_max_count = consume_max_count

        super().__init__(app_function, consume_topics_list, produce_topic_schema_dict=produce_topic_schema_dict,
                         transaction_type=transaction_type,
                         app_function_arglist=app_function_arglist, metrics_manager=metrics_manager,
                         schema_registry=schema_registry, cluster_name=cluster_name, consumer=consumer,
                         producer=producer)

    def prepare_transaction(self, *args, **kwargs):
        return super().consume(*args, **kwargs)

    def _max_consume_time_continue(self):
        if self._time_elapse_max_seconds:
            seconds_elapsed = datetime.now().timestamp() - self._time_elapse_start
            return seconds_elapsed < self._time_elapse_max_seconds
        return True

    def _max_consume_count_continue(self):
        if self._consume_max_count:
            return self.transaction.message_count < self._consume_max_count
        return True

    def _keep_consuming(self):
        return self._max_consume_count_continue() and self._max_consume_time_continue()

    def commit(self):
        # TODO: remove once base class one is changed back to this
        if not self.transaction._committed:
            self.transaction.commit()

    def _app_run(self, *args, **kwargs):
        self.prepare_transaction(*args, **kwargs)
        if self._time_elapse_max_seconds:
            self._time_elapse_start = datetime.now().timestamp()
        LOGGER.info(
            f'Processing up to {self._consume_max_count} messages for up to {self._time_elapse_max_seconds} seconds!')
        try:
            if not self.transaction._committed:
                while self._keep_consuming():
                    self.transaction.consume()
                    self.app_function(self.transaction, *self.app_function_arglist)
        except NoMessageError:
            self.producer.poll(0)
            LOGGER.info('No messages!')
            pass
        if self.transaction._commit_offsets:
            self.commit()
